from utils.BinTree import BinTree

import gtk


def _cmp_equal(this, other):

    return cmp(this, other)


def _cmp_notequal(this, other):

    return cmp(this, other) or 1


def _cmp_snap(this, other):

    u, ua, ub = this
    v, va, vb = other

    if (abs(u - v) < 8):
        if (va <= ua <= vb or va <= ub <= vb or
            ua <= va <= ub or ua <= vb <= ub):
            return 0

    if (u < v): return -1
    else: return 1


#
# Class for snapping windows together.
#
class WindowSnapper(object):

    __slots__ = ('__horiz', '__vertic')

    def __init__(self):

        self.__horiz = BinTree()
        self.__vertic = BinTree()

        self.__horiz.insert((0, 0, gtk.gdk.screen_height()), _cmp_notequal)
        self.__horiz.insert((gtk.gdk.screen_width(),
                             0, gtk.gdk.screen_height()), _cmp_notequal)
        self.__vertic.insert((0, 0, gtk.gdk.screen_width()), _cmp_notequal)
        self.__vertic.insert((gtk.gdk.screen_height(),
                             0, gtk.gdk.screen_width()), _cmp_notequal)


    def insert(self, x, y, w, h):

        self.__horiz.insert((x, y, y + h), _cmp_notequal)
        self.__horiz.insert((x + w, y, y + h), _cmp_notequal)
        self.__vertic.insert((y, x, x + w), _cmp_notequal)
        self.__vertic.insert((y + h, x, x + w), _cmp_notequal)


    def remove(self, x, y, w, h):

        self.__horiz.remove((x, y, y + h), _cmp_equal)
        self.__horiz.remove((x + w, y, y + h), _cmp_equal)
        self.__vertic.remove((y, x, x + w), _cmp_equal)
        self.__vertic.remove((y + h, x, x + w), _cmp_equal)


    def snap(self, x, y, w, h):

        value1 = self.__horiz.find((x, y, y + h), _cmp_snap)
        value2 = self.__horiz.find((x + w, y, y + h), _cmp_snap)
        value3 = self.__vertic.find((y, x, x + w), _cmp_snap)
        value4 = self.__vertic.find((y + h, x, x + w), _cmp_snap)

        new_x = x
        new_y = y
        if (value1):
            new_x = value1[0]
        elif (value2):
            new_x2 = value2[0]
            new_x = new_x2 - w
        if (value3):
            new_y = value3[0]
        elif (value4):
            new_y2 = value4[0]
            new_y = new_y2 - h

        return (new_x, new_y)



def create_arch():

    arch = __detect_arch()
    arch._init()
    return arch


def __detect_arch():

    if (HAVE_WIN32):
        import Windows
        return Windows.Win32()

    import os
    uname = os.uname()

    if (uname[0] == 'Linux'):

        import Linux

        if (uname[-1] in ('i386', 'i486', 'i586', 'i686', 'athlon', 'x86_64')):
            return Linux.X86()

        if (uname[-1] in ('sparc', 'sparc64')):
            return Linux.Sparc()

        if (uname[-1] in ('ppc', 'ppc64')):
            return Linux.PPC()

        return Linux.Generic()


    elif (uname[0] == 'FreeBSD'):

        import FreeBSD

        return FreeBSD.Generic()


    elif (uname[0] == 'OpenBSD'):

        import OpenBSD

        return OpenBSD.Generic()


    elif (uname[0] == 'NetBSD'):

        import NetBSD

        return NetBSD.Generic()


    log("OS/Architecture not found!")

    import Arch

    return Arch.Arch()




from ConfigWidget import ConfigWidget
from utils.datatypes import *
from layout import Unit

import gtk

class ConfigUnit(ConfigWidget):

    def __init__(self, name, getter, setter, caller):

        self.__lock = False

        self.__units = []
        self.__prev_unit = Unit.UNIT_PX


        ConfigWidget.__init__(self, name, getter, setter, caller)

        self._register_property("min", TYPE_FLOAT, self._setp_lhd,
                                self._getp, 0.0, doc = "Lower bound")
        self._register_property("max", TYPE_FLOAT, self._setp_lhd,
                                self._getp, 9999.0, doc = "Upper bound")
        self._register_property("digits", TYPE_INT, self._setp_lhd,
                                self._getp, 2,
                                doc = "Numbers of decimal digits")
        self._register_property("increment", TYPE_FLOAT, self._setp_increment,
                                self._getp, 1, doc = "Size of increments")
        self._register_property("value", TYPE_UNIT, self._setp_value,
                                self._getp, Unit.ZERO, doc = "Value")



    def get_widgets(self):

        low = self.get_prop("min")
        high = self.get_prop("max")
        digits = self.get_prop("digits")

        self.__label = gtk.Label("")
        self.__label.show()
        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.show()
        align.add(self.__label)

        box = gtk.HBox(spacing = 6)
        box.show()

        adjustment = gtk.Adjustment(0, -0xffffff, 0xffffff, 1, 1, 0)
        self.__spin_button = gtk.SpinButton(adjustment, 1, digits)
        self.__spin_button.set_numeric(True)
        self.__spin_button.show()
        self.__spin_button.connect("value-changed", self.__on_change)
        box.pack_start(self.__spin_button, True, True)

        self.__optmenu = gtk.combo_box_new_text()
        self.__optmenu.show()
        for unit, name in ((Unit.UNIT_PX, _("pixel")),
                           (Unit.UNIT_PT, _("point")),
                           (Unit.UNIT_CM, _("cm")),
                           (Unit.UNIT_IN, _("inch")),
                           (Unit.UNIT_PERCENT, "%")):
            self.__units.append(unit)
            self.__optmenu.append_text(name)
        #end for
        self.__optmenu.set_active(0)

        self.__optmenu.connect("changed", self.__on_change_unit)
        box.pack_end(self.__optmenu, False, False)

        return (align, box)


    def __on_change(self, src):

        size = self.__spin_button.get_value()
        unit = self.__units[self.__optmenu.get_active()]

        new_value = Unit.Unit(size, unit)
        self._setp("value", new_value)
        self._set_config(Unit.Unit(size, unit))


    def __on_change_unit(self, src):

        if (self.__lock): return
        
        unit = self.__units[src.get_active()]
        value = self.get_prop("value")
        
        if (unit == Unit.UNIT_PX): size = value.as_px()
        elif (unit == Unit.UNIT_PT): size = value.as_pt()
        elif (unit == Unit.UNIT_CM): size = value.as_cm()
        elif (unit == Unit.UNIT_IN): size = value.as_in()
        elif (unit == Unit.UNIT_PERCENT): size = value.as_percent()

        self.__spin_button.set_value(size)


    def _set_enabled(self, value): self.__spin_button.set_sensitive(value)
    def _set_label(self, value): self.__label.set_text(value)


    def _setp_lhd(self, key, value):

        low, high = self.__spin_button.get_range()
        if (key == "min"):
            self.__spin_button.set_range(value, high)
        elif (key == "max"):
            self.__spin_button.set_range(low, value)
        elif (key == "digits"):
            self.__spin_button.set_digits(value)

        self._setp(key, value)


    def _setp_increment(self, key, value):

        self.__spin_button.set_increments(value, value * 10);
        self._setp(key, value)



    def _setp_value(self, key, value):

        unit = value.get_unit()
        if (unit == Unit.UNIT_PX): size = value.as_px()
        elif (unit == Unit.UNIT_PT): size = value.as_pt()
        elif (unit == Unit.UNIT_CM): size = value.as_cm()
        elif (unit == Unit.UNIT_IN): size = value.as_in()
        elif (unit == Unit.UNIT_PERCENT): size = value.as_percent()

        self.__lock = True
        self.__optmenu.set_active(self.__units.index(unit))
        self.__spin_button.set_value(size)
        self.__lock = False
        self._setp(key, value)



from main import UNSET_COORD

import gtk


#
# Container class for displays that are pluggable into other applications.
#
class Plug(gtk.Plug):

    def __init__(self):

        # window size for detecting resizing
        self.__window_size = (0, 0)

        self.__shape = None

        gtk.Plug.__init__(self, 0L)

        self.set_size_request(-1, -1)
        self.set_default_size(10, 10)

        # set up event handlers
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK |
                        gtk.gdk.BUTTON_RELEASE_MASK)



    #
    # Returns the XEMBED ID for embedding this plug into other applications.
    #
    def get_xembed_id(self):

        return self.get_id()



    def close(self):

        self.destroy()



    def set_position(self, x, y):

        # we happily ignore this
        pass



    def set_size(self, width, height):

        self.resize(width, height)



    def set_window_flags(self, value):

        # we happily ignore this
        pass



    def set_shape(self, mask):

        if (self.__window_pos == (UNSET_COORD, UNSET_COORD)):
            self.__shape = mask
        else:
            self.shape_combine_mask(mask, 0, 0)



#
# Simple class for modifiable and easy to use menu items.
#
class MenuItem(object):

    __slots__ = ("path", "label", "icon", "callback", "args", "active")


    def __init__(self, path, label = None, icon = None,
                 callback = None, args = [], active = True):

        self.path = path
        self.label = label
        self.icon = icon
        self.callback = callback
        self.args = args
        self.active = active



from shell.Plugin import Plugin
from utils.Observable import Observable
from utils.Trie import Trie
from main import DISPLAYPATHS
from main.MetaData import MetaData
from utils.QueryParser import QueryParser
from utils.Struct import Struct

import os


#
# Plugin for the collection of installed displays.
#
class Core_DisplayCollection(Plugin, Observable):

    OBS_RELOAD = 0

    def init(self):
        self.__trie = None
        self.__queryparser = None

        self.reload()


    def __find_displays(self, path):

        files = os.listdir(path)
        for f in files:
            filepath = os.path.join(path, f)
            if (os.path.isdir(filepath)):
                self.__find_displays(filepath)
            else:
                name, ext = os.path.splitext(f)
                if (ext == ".display"):
                    meta = MetaData(filepath)
                    item = Struct(path = filepath,
                                  name = meta.get(meta.KEY_NAME),
                                  version = meta.get(meta.KEY_VERSION),
                                  author = meta.get(meta.KEY_AUTHOR),
                                  description = meta.get(meta.KEY_DESCRIPTION),
                                  category = meta.get(meta.KEY_CATEGORY),
                                  preview = meta.get(meta.KEY_PREVIEW))

                    self.__insert_into_trie(item)
        #end for


    def __insert_into_trie(self, item):

        self.__trie.insert(["name"] + list(item.name), item)
        self.__trie.insert(["path"] + list(item.path), item)
        self.__trie.insert(["author"] + list(item.author), item)
        self.__trie.insert(["version"] + list(item.version), item)
        self.__trie.insert(["description"] + list(item.description),
                           item)
        for c in item.category.split(","):
            self.__trie.insert(["category"] + list(c.strip()), item)


    def add_observer(self, obs):

        def f(src, cmd): obs()
        Observable.add_observer(self, f)
        

    def query(self, query):

        matches = self.__queryparser.parse(query)
        return matches


    def remove(self, item):

        self.__trie.remove(item)
        self.update_observer(self.OBS_RELOAD)


    def reload(self):

        self.__trie = Trie()
        self.__trie.set_case_sensitive(False)
        self.__queryparser = QueryParser(self.__trie)

        # read the display files
        for path in DISPLAYPATHS:
            if (os.path.isdir(path)):
                self.__find_displays(path)
        self.update_observer(self.OBS_RELOAD)
        

def get_class(): return Core_DisplayCollection



from utils import vfs

import os


class MetaData:

    """
    Class for reading and manipulating meta info of .display files.
    """

    # the valid meta keys
    KEY_AUTHOR = "author"
    KEY_NAME = "name"
    KEY_VERSION = "version"
    KEY_DESCRIPTION = "description"
    KEY_CATEGORY = "category"
    KEY_ICON = "icon"
    KEY_PREVIEW = "preview"


    def __init__(self, displayfile):

        self.__file = displayfile
        self.__directory = ""
        self.__filename = ""
        self.__values = {}


        self.__directory, self.__filename = os.path.split(displayfile)
        self.__read_meta()



    def __find_meta(self, data):

        """
        Finds the meta information tag in the given data and returns a tuple of
        three strings: the part before the meta info, the meta info, and the
        part behind the meta info.
        """

        # states are: 0 find "<"
        #             1 read comment
        #             2 read meta
        state = 0

        position = 0
        limit = len(data)
        while (position < limit):
            # check every "<"
            if (state == 0):
                index = data.find("<", position)
                if (index == -1):
                    break
                position = index + 1
                if (data[index:].startswith("<!--")):
                    state = 1
                elif (data[index:].startswith("<meta")):
                    state = 2

            # ignore comments
            elif (state == 1):
                index = data.find("-->", position)
                if (index == -1):
                    break
                position = index
                state = 0

            # divide into premeta, meta, postmeta parts
            elif (state == 2):
                startpos = position + len("<meta")
                endpos = data.find(">", position)
                if (endpos == -1):
                    break
                endpos = data[:endpos].rfind("/")
                return (data[:startpos], data[startpos:endpos], data[endpos:])


        return (data, "", "")


    def  __parse_meta(self, meta):

        """
        Parses the given meta data string and returns a dictionary with keys
        and values.
        """

        # states are: 0 read key name
        #             1 read " string
        #             2 read ' string
        state = 0

        position = 0
        limit = len(meta)
        key = ""
        dikt = {}
        while (position < limit):
            # read key names
            if (state == 0):
                index = meta.find("=", position)
                if (index == -1):
                    break
                key = meta[position:index].strip()
                if (meta[index + 1].strip().startswith("\"")):
                    state = 1
                elif (meta[index + 1].strip().startswith("'")):
                    state = 2
                position = index

            # read values quoted with "
            elif (state == 1):
                position = meta.find("\"", position)
                index = meta.find("\"", position + 1)
                if (index == -1):
                    break
                value = meta[position + 1:index]
                dikt[key] = value
                position = index + 1
                state = 0

            # read values quoted with '
            elif (state == 2):
                position = meta.find("'", position)
                index = meta.find("'", position + 1)
                if (index == -1):
                    break
                value = meta[position + 1:index]
                dikt[key] = value
                position = index + 1
                state = 0


        return dikt



    def __read_meta(self):

        """ Reads the meta information in the file.  """

        data = vfs.read_entire_file(self.__file) #open(self.__file).read()

        meta = self.__find_meta(data)[1]
        if (meta):
            self.__values = self.__parse_meta(meta)

        if self.KEY_NAME not in self.__values:
            self.__values[self.KEY_NAME] = \
              os.path.splitext(os.path.basename(self.__file))[0]



    def get_directory(self):

        """ Returns the directory of the display. """
        return self.__directory



    def get_filename(self):

        """ Returns the filename of the display. """

        return self.__filename



    def get(self, key):

        """ Returns the value of the given meta key. """

        return self.__values.get(key, "")



    def __cmp__(self, other):

        """
        Compares two MetaData objects and makes MetaData objects thus sortable.
        """

        return cmp(self.get(self.KEY_NAME), other.get(other.KEY_NAME))



"""
Class for a simple mini DOM.

This class takes an SVG string at construction time and then provides a way to
modify element properties in the DOM tree.
The DOM object can be serialized to a XML string at any time.
"""

from xml import sax
    

class DOM(sax.handler.ContentHandler):

    def __init__(self, xml):

        self.__dom = None
        self.__node_stack = []
        self.__current_node = None
        self.__current_chars = ""
        self.__id_table = {}

        sax.handler.ContentHandler.__init__(self)

        # silly PyXML fails for standalone="no" documents without net access,
        # so rip that out
        xml = self.__modify_standalone(xml)

        try:
            sax.parseString(xml, self)
        except sax._exceptions.SAXParseException, exc:
            raise UserError(_("XML parse error"),
                            _("An invalid graphics code was loaded into a "
                              "canvas."),
                            code = xml, lineno = exc.getLineNumber())
        except:
            return

        self.__dom.set_id_table(self.__id_table)


    #
    # Rips out a standalone="yes" attribute if it's set. We don't need and don't
    # want it for SVGs.
    #
    def __modify_standalone(self, xml):

        index1 = xml.find("<")

        if (xml[index1 + 1] != "?"):
            # no header; nothing to do
            return xml

        index2 = xml.find("?>")
        header = xml[index1:index2]
        index3 = header.find("standalone")

        if (index3 == -1):
            # no "standalone setting; add one
            return xml[:index1] + " standalone='yes'" + xml[index2:]

        state = 0
        index4 = index3 + 10
        for c in header[index4:]:
            index4 += 1
            if (c in ("\"", "'")): state += 1
            if (state == 2): break

        if ("no" in header.lower()[index3:index4]):
            # replace with "yes"
            xml = xml[:index1 + index3] + \
        xml[index1 + index3:index1 + index4].lower().replace("no", "yes") + \
                  xml[index1 + index4:]
            
        return xml
    

    def get_root(self): return self.__dom
    

    def startElement(self, name, attrs):

        parent = self.__current_node
        self.__current_node = _DOMNode(name)
        self.__current_chars = ""
        for key, value in attrs.items():
            self.__current_node[key] = value

        if (parent): parent.add_child(self.__current_node)
        else:
            # we insert an empty <g> node here; it can be used later if needed
            # TODO: not here; this is meant to be a generic DOM
            self.__dom = self.__current_node
            new_node = _DOMNode("g")
            self.__dom.add_child(new_node)
            self.__current_node = new_node

        if ("id" in attrs.keys()):
            self.__id_table[attrs["id"]] = self.__current_node

        self.__node_stack.append(self.__current_node)
        

    def endElement(self, name):

        if (self.__current_chars):
            self.__current_node["PCDATA"] = self.__current_chars

        self.__node_stack.pop(-1)
        if (self.__node_stack):
            self.__current_node = self.__node_stack[-1]


    def characters(self, content):

        self.__current_chars += content



#
# Class for nodes of the DOM tree.
#
class _DOMNode:

    def __init__(self, name):

        self.__name = name
        self.__attrs = {}
        self.__children = []
        self.__update_handler = None
        self.__id_table = {}


    #
    # Sets the ID table for quickly accessing a node by its ID.
    #
    def set_id_table(self, table): self.__id_table = table


    #
    # Returns whether the given key exists.
    #
    def has_key(self, key): return (key in self.__attrs)


    #
    # Returns the node with the given ID.
    #
    def get(self, ident):

        try:
            return self.__id_table[ident]
        except KeyError:
            raise UserError(_("No such element: %s") % ident,
                           _("The element with ID <b>%s</b> does not "
                             "exist in the SVG image.") % ident)


    #
    # Sets the handler for updating the tree.
    #
    def set_update_handler(self, handler):

        self.__update_handler = handler

    #
    # Updates the tree.
    #
    def update(self):

        if (self.__update_handler): self.__update_handler()
        

    def add_child(self, child): self.__children.append(child)
    def get_children(self): return self.__children[:]


    def __getitem__(self, key):

        try:
            return self.__attrs[key]
        except KeyError:
            raise UserError(_("No such property: %s") % key,
                           _("The SVG element <b>%s</b> does not have the "
                             "<b>%s</b> property.") % (self.__name, key))
        
    def __setitem__(self, key, value): self.__attrs[key] = value


    #
    # Returns a XML representation of this node.
    #
    def __str__(self):

        attrs = [ "%s=\"%s\"" % (k, v) for k, v in self.__attrs.items() ]
        attrs = " ".join(attrs)
        children = [ str(c) for c in self.get_children() ]
        children = "\n".join(children)
        contents = self.__attrs.get("PCDATA", "")
        contents += children
        if (contents):
            out = "<%s %s>%s</%s>" % (self.__name, attrs, contents, self.__name)
        else:
            out = "<%s %s/>" % (self.__name, attrs)

        return out



from ConfigWidget import ConfigWidget
from utils.datatypes import *

import gtk


class ConfigFont(ConfigWidget):

    def __init__(self, name, getter, setter, caller):

        ConfigWidget.__init__(self, name, getter, setter, caller)

        self._register_property("value", TYPE_STRING, self._setp_value,
                                self._getp, False, doc = "Font description")


    def get_widgets(self):

        self.__label = gtk.Label("")
        self.__label.show()
        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.show()
        align.add(self.__label)

        self.__btn = gtk.FontButton()
        self.__btn.set_use_font(True)
        self.__btn.set_use_size(False)
        self.__btn.set_show_style(True)
        self.__btn.set_show_size(True)
        self.__btn.show()
        self.__btn.connect("font-set", self.__on_change)

        return (align, self.__btn)


    def __on_change(self, src):

        value = src.get_font_name()
        self._set_config(value)


    def _set_enabled(self, value): self.__btn.set_sensitive(value)
    def _set_label(self, value): self.__label.set_text(value)

    def _setp_value(self, key, value):

        self.__btn.set_font_name(value)

        # force "font-set" signal
        self.__btn.emit("font-set")
        
        self._setp(key, value)
        



from libdesklets.system.Arch import Arch

import re
import struct


class Generic(Arch):

    def __init__(self):

        Arch.__init__(self)

        self.__bogomips    = re.compile('^bogomips\s+:\s+(\d+\.\d+)$', re.M)
        self.__net_devices = re.compile('^\s*(\w+):', re.M)


    def _read_cpuinfo(self):
        """
        @return : content of the pseudo file /proc/cpuinfo
        @rtype  : str
        """

        return open('/proc/cpuinfo').read()



    def net_devices(self):
        """
        @return : all available network devices
        @rtype  : list
        """

        data = open('/proc/net/dev').read()
        return self.__net_devices.findall(data)



    def cpu_bogomips(self):
        """
        @return : bogomips of installed processor
        @rtype  : float
        """

        m = self.__bogomips.search( self._read_cpuinfo() )
        return float(m.group(1))



    def users(self):
        """
        @return : number of connected users
        @rtype  : int
        """

        # man utmp
        # don't know if
        # sizeof(struct utmp) == 384
        # sizeof(short) == 2
        # on every Linux arch
        # http://gnomesupport.org/forums/viewtopic.php?p=33686
        # X86, X86_64 : ok

        count = 0
        data = open('/var/run/utmp', 'rb').read()

        for i in range(0, len(data), 384):

            ut_type = struct.unpack('h', data[i:i+2])[0]

            if (ut_type == 7):
                count += 1

        return count



import gtop
from Device import Device

import threading


class Total(Device):

    def __init__(self):

        Device.__init__(self)

        self.__total = 0.0
        self.__user  = 0.0
        self.__sys   = 0.0
        self.__nice  = 0.0
        self.__idle  = 0.0
        self.__lock  = threading.Lock()



    def _get_cpu(self): return gtop.cpu()


    def poll(self):

        """
        @return : cpu load
        @rtype  : float
        """
        self.__lock.acquire()

        try:
            cpu = self._get_cpu()
            totaldiff = cpu.total - self.__total
            load = float(cpu.frequency * ((cpu.user + cpu.sys + cpu.nice) -
                                          (self.__user + self.__sys + self.__nice))
                         / (totaldiff + 0.001) + 0.5)

            if (load > 100.0): load = 100.0

            self.__total = cpu.total
            self.__user  = cpu.user
            self.__sys   = cpu.sys
            self.__nice  = cpu.nice
            self.__idle  = cpu.idle

            return load

        finally:
            self.__lock.release()




class CPU(Total):

    def __init__(self, n):

        Total.__init__(self)
        self.__n = n


    def _get_cpu(self): return gtop.cpu().cpus[self.__n]



import socket
import re
import sys

from utils.Struct import Struct
from libdesklets import convert


__HOST = 'localhost'
__PORT = 7634
# __SEPARATOR = '|'
__REGEX = re.compile('\|(?P<device>.+?)\|(?P<name>.+?)\|(?P<value>\d+)\|(?P<unit>[CF])\|')


def __split(data):
    
    # |/dev/hda|TOSHIBA MK6025GAS|50|C|
    # |/dev/hda|MAXTOR 6L040J2|41|C||/dev/hdh|IC35L040AVVN07-0|39|C|
    
    def match_to_struct(m):

        value = float(m['value'])
        unit  = m['unit']
        
        assert unit in ('C', 'F')

        if   unit == 'C':
            C = value
        elif unit == 'F':
            C = convert.fahrenheit_to_centigrade(value)
        
        F = convert.centigrade_to_fahrenheit(C)
        K = convert.centigrade_to_kelvin(C)

        return Struct(device = m['device'],
                      name   = m['name'],
                      temp   = Struct(centigrade = C,
                                      fahrenheit = F,
                                      kelvin     = K
                                      )
                      )

    return [ match_to_struct(m.groupdict()) for m in __REGEX.finditer(data) ]



def __socket_read():    

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((__HOST, __PORT))
    return s.makefile().read()




def poll_all():
    
    try:
        return __split( __socket_read() )
    except Exception, exc:
        log("Cannot retrieve HDD temperature. Error was %s" % (exc,))
        return []



def available_devices():
    
    return [ s.device for s in poll_all() ]



def poll(device):

    for s in poll_all():

        if s.device == device:
            return s

    else:
        log("Cannot retrieve HDD temperature for device %s." % (device,))
        return Null




if __name__ == '__main__':

    print available_devices()
    print poll('/dev/hda')
    print poll_all()
    print poll('/dev/uba')




from main import HOME

import gtk
import gobject



#
# Class for a list of items.
#
class ListView(gtk.TreeView):

    def __init__(self, columntypes, renderer):

        self.__items = {}
        self.__renderer = renderer


        gtk.TreeView.__init__(self)
        self.__liststore = gtk.ListStore(*columntypes)

        self.set_model(self.__liststore)
        self.set_headers_visible(False)

        cnt = 0
        for ctype in columntypes:
            if (ctype == gtk.gdk.Pixbuf):
                col = gtk.TreeViewColumn(None, gtk.CellRendererPixbuf(),
                                         pixbuf = cnt)
            elif (ctype == gobject.TYPE_STRING):
                col = gtk.TreeViewColumn(None, gtk.CellRendererText(),
                                         markup = cnt)
            
            self.append_column(col)
            cnt += 1
        #end for
            



    #
    # Clears the list.
    #
    def clear(self):

        self.__items = {}
        self.__liststore.clear()
        self.set_size_request(-1, -1)
        import gc; gc.collect()


    #
    # Returns the selected item or None.
    #
    def get_selected_item(self):

        path = self.get_cursor()[0]
        return self.__items.get(path)


    #
    # Selects the given item.
    #
    def select_item(self, item):

        for path, i in self.__items.items():
            if (item == i):
                self.set_cursor(path, None, False)
                break
        #end for



    #
    # Adds as item to the list.
    #
    def add_item(self, item):

        values = self.__renderer(item)

        iter = self.__liststore.append()
        for i in range(len(values)):
            self.__liststore.set(iter, i, values[i])

        path = self.__liststore.get_path(iter)
        self.__items[path] = item



import Permission
from main import _

import types


class Interface(object):

    """Abstract base class for interfaces."""


    def __get_members(impl):

        """Simply returns all members found in a class."""

        return [ (name, member) for name, member in impl.__dict__.items() ]

    __get_members = staticmethod(__get_members)



    def get_properties(iface):
        
        """Returns a list of the properties of the given implementation."""

        return [ (n, p) for n, p in Interface.__get_members(iface)
                 if (type(p) == property) ]

    get_properties = staticmethod(get_properties)



    def get_methods(iface):
        
        """Returns a list of the methods of the given implementation."""

        return [ (n, "") for n, m in Interface.__get_members(iface)
                if (type(m) == types.FunctionType) ] # or MethodType ?

    get_methods = staticmethod(get_methods)



    def get_permissions(iface):

        """Returns a list of the permissions of the properties."""

        perms = [ (n, perm) for n, perm in Interface.__get_members(iface)
                  if (isinstance(perm, Permission._Permission)) ]

        return perms
    
    get_permissions = staticmethod(get_permissions)



    def get_taz_style_id(iface):
        """
          Computes and returns the interface in the wrong way for maintaining
          backwards compatibility.
        """

        import md5
        import utils

        name = iface.__name__
        ash = md5.new(name + ":").hexdigest()
        # encode it in base36 to get shorter ID
        ash = utils.radix( int(ash, 16), 36 )
        # ID have to be 25 char long
        ash = '0' * (25 - len(ash)) + ash
        assert (len(ash) == 25)
        return ("%s:%s" % (name, ash))

    get_taz_style_id = staticmethod(get_taz_style_id)
        


    def get_id(iface):

        """Returns the unique identifier of the given interface."""

        # the algorithm is simple:
        # build a string containing the class name and a sorted list of the
        # properties and return the MD5 fingerprint of it
        import md5
        import utils

        name = iface.__name__
        items1 = Interface.get_permissions(iface)
        items1.sort()
        items2 = Interface.get_methods(iface)
        items2.sort()

        tmp = ["%s:%s" % (k, v) for k, v in items1 + items2]
        tmp = "%s:%s" % (name, ",".join(tmp))

        ash = md5.new(tmp).hexdigest()
        # encode it in base36 to get shorter ID
        ash = utils.radix( int(ash, 16), 36 )
        # ID have to be 25 char long
        ash = '0' * (25 - len(ash)) + ash

        assert (len(ash) == 25)

        # the "-2" suffix is for marking the ID as rev 2
        iface_id = "%s:%s-2" % (name, ash)
        return iface_id

    get_id = staticmethod(get_id)


    def get_rev_of_id(ident):

        """Returns the revision of the format in which the ID is computed."""

        if (ident[-2] != "-"): return "1"
        else: return ident[-1]

    get_rev_of_id = staticmethod(get_rev_of_id)



    def get_interfaces(impl):

        """Returns the interfaces implemented by the given class"""

        return [ c for c in impl.mro()
                 if issubclass(c, Interface) and not c in (Interface, impl) ]

    get_interfaces = staticmethod(get_interfaces)



    def assert_interfaces(impl):
        """Static method for asserting that the interfaces are implemented
           properly in the given class. Throws exceptions otherwise."""

        # find the implemented properties
        impl_props = dict(impl.get_properties())

        # find the implemented methods
        impl_meths = dict(impl.get_methods())

        # check all interfaces
        for iface in Interface.get_interfaces(impl):

            # check properties
            required = iface.get_permissions()
            for name, perm in required:
                if name not in impl_props:
                    raise NotImplementedError("Missing implementation for "
                                              "%s.%s" % (iface.__name__, name))

                # check if the property's mode is according to the interfaces
                if (perm != Permission.Permission(implemented[name])):
                    raise NotImplementedError("Broken implementation for "
                                      "%s.%s" % (iface.__name__, name))
            #end for

            # check methods
            required = iface.get_methods()
            for name, meth in required:
                if name not in impl_meths:
                    raise NotImplementedError("Missing implementation for "
                                              "%s.%s" % (iface.__name__, name))
            #end for

        #end for

    assert_interfaces = staticmethod(assert_interfaces)



    def text_describe(impl):

        """Returns a description of the interfaces of the given class."""

        textlist = []
        interfaces = Interface.get_interfaces(impl)

        for i in interfaces:

            out = ""

            ident = Interface.get_id(i)

            out += ident + '\n\n'

            for key in dir(i):

                value = impl.__dict__.get(key)
                if (not value): continue

                if (isinstance(value, property)):
                    out += "  "
                    out += key.ljust(25)
                    perm = str(Permission.Permission(value)).ljust(4)
                    out += perm
                    out += "%s \n" % (value.__doc__ or _("no description"), )

            out += "\n"

            textlist.append(out)

        return ''.join(textlist)

    text_describe = staticmethod(text_describe)


    def gui_describe(impl):

        guilist = []
        interfaces = Interface.get_interfaces(impl)

        for i in interfaces:

            items = []
            
            for key in dir(i):

                value = impl.__dict__.get(key)
                if (not value): continue

                if (isinstance(value, property)):
                    items.append((key, str(Permission.Permission(value)),
                                    value.__doc__ or _("no description")))

            guilist.append( (Interface.get_id(i), items) )

        return guilist


    gui_describe = staticmethod(gui_describe)




from shell.Plugin import Plugin

import gtk


class Shell_StatusBar(Plugin):

    def init(self):
        bar = gtk.Statusbar()
        bar.show()
        self.__widget = bar

        shell = self._get_plugin("UI_Shell")
        shell.set_statusbar(self.__widget)

        

    def set_status(self, text):

        self.__widget.push(self.__widget.get_context_id("message"), text)


def get_class(): return Shell_StatusBar



from main import DISPLAY, SOCKET_PATH
from main.remotecommands import *
from utils import xdr

import atexit
import os
import sys
import socket
import gobject


class RemoteSocket:
    """
      This class implements a socket server for accepting incoming commands
      from clients.
      Socket server programming not necessarily requires threads. Actually, by
      not using threads, we gain some performance and stability.
    """

    def __init__(self):

        # handlers for incoming messages
        self.__message_handlers = {}

        self.__stopevent = False

        # marks the socket as blocked, i.e. must not accept new requests
        self.__block_timeout = {}



    def start(self):
        """
          Starts the server. Call this after the handlers have been set.
        """

        # create the socket
        if (HAVE_WIN32):
            serversock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sockfile = ""  # irrelevant
            try:
                serversock.bind(("localhost", 1979))
            except socket.error:
                log("Couldn't bind to socket. Aborting.")
                sys.exit(1)

        else:
            import utils
            utils.makedirs(SOCKET_PATH)
            serversock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sockfile = os.path.join(SOCKET_PATH, DISPLAY)
            try:
                serversock.bind(sockfile)
            except socket.error:
                try:
                    os.remove(sockfile)
                except OSError:
                    log("Couldn't remove dead socket file \"%s\"." % sockfile)
                    sys.exit(1)
                try:
                    serversock.bind(sockfile)
                except socket.error:
                    log("Couldn't bind to socket. Aborting.")
                    sys.exit(1)

        serversock.listen(3)
        serversock.setblocking(False)

        atexit.register(self.__shutdown, serversock, sockfile)

        gobject.timeout_add(100, self.__server_handler, serversock, sockfile)



    def add_message_handler(self, message, handler):
        """
          Adds a new message handler.
        """

        self.__message_handlers[message] = handler



    def __shutdown(self, serversock, sockfile):
        """
          Shuts down the server and closes/removes the used socket.
        """

        self.__stopevent = True
        serversock.close()
        if (sockfile):
            os.remove(sockfile)



    def __server_handler(self, serversock, sockfile):
        """
          The socket server reactor.
        """

        if (self.__stopevent):
            return False

        try:
            cnx, addr = serversock.accept()
        except socket.error:
            pass

        else:
            # new connection
            cnx.setblocking(False)
            self.__block_timeout[cnx] = 0
            gobject.timeout_add(250, self.__client_handler, cnx)

        return True



    def __client_handler(self, clientsock):
        """
          The client server reactor. A new client gets its own new reactor.
        """


        def callback(*retval):

            self.__block_timeout[clientsock] = 0

            try:
                xdr.send(clientsock, *retval)

            # hmm, we aren't able to send somthing here
            except socket.error, exc:
                log("Socket is dead!\n%s\n" % exc)
                return False

            # the data seems to be of size zero here
            except xdr.XDRError, exc:
                log(`exc`)

            # something really unexpected has happened
            except:
                try:
                    xdr.send_error(clientsock)
                except:
                    import traceback; traceback.print_exc()


        if (self.__block_timeout[clientsock] > 0):
            self.__block_timeout[clientsock] -= 1

            if (self.__block_timeout[clientsock] == 0):
                callback()
            return True

        try:
            data = xdr.recv(clientsock)

        # no data on connected socket
        except socket.error:
            return True

        # socket disconnected
        except xdr.XDRError:
            clientsock.close()
            return False

        message = data[0]
        args = data[1:]

        try:
            handler = self.__message_handlers[message]

        # message_handler doesn't exist
        except KeyError:
            log("Key %s doesn't exist in message_handlers!" % (message,))
            clientsock.close()
            return False

        # timeout after 20 * 250 ms
        self.__block_timeout[clientsock] = 20
        handler(callback, *args)
        return True

        #
        #
        #

        try:
            #handler(callback, *args)
            retval = handler(*args)
            if (retval == None): retval = ()

        # something bad happened while trying to execute the handler
        except:
            from utils.ErrorFormatter import ErrorFormatter
            details = ErrorFormatter().format(sys.exc_info())
            log("Execution of handler (%s, %s) failed!\n%s" % (`handler`,
                                                               `args`,
                                                               details))
            # here we get our broken pipes
            retval = ()

        try:
            xdr.send(clientsock, *retval)

        # hmm, we aren't able to send somthing here
        except socket.error, exc:
            log("Socket is dead!\n%s\n" % exc)
            return False

        # the data seems to be of size zero here
        except xdr.XDRError, exc:
            log(`exc`)

        # something really unexpected has happened
        except:
            try:
                xdr.send_error(clientsock)
            except:
                import traceback; traceback.print_exc()

        return True



from libdesklets.controls import Control, FakeSelf
from libdesklets import system

from ISystem import ISystem


class System(Control, ISystem):

    def __init__(self):

        Control.__init__(self)

        self.__allfs = False
        self.__dev   = ""
        self.__mdir  = ""
        self.__pid   = 0
        self.__which = 0



    def __get_allfs(self):

        return self.__allfs

    def __set_allfs(self, bval):

        self.__allfs = bval
        self._update("allfs")



    def __get_mdir(self):

        return self.__mdir

    def __set_mdir(self, val):

        self.__mdir = val
        self._update("mountdir")



    def __get_dev(self):

        return self.__dev

    def __set_dev(self, val):

        self.__dev = val
        self._update("device")



    def __get_pid(self):

        return self.__pid

    def __set_pid(self, val):

        self.__pid = val
        self._update("pid")


    def __get_which(self):

        return self.__which

    def __set_which(self, val):

        self.__which = val
        self._update("which")


    def __get_fsusage(self):

        return system.fsusage(self.__mdir)

    def __get_net_load(self):

        return system.net_load(self.__dev)

    def __get_net_speed(self):

        return system.net_speed(self.__dev)

    def __get_net_state(self):

        return system.net_state(self.__dev)

    def __get_ppp(self):

        return system.ppp(self.__dev)

    def __get_proc_args(self):

        return system.proc_args(self.__pid)

    def __get_proc_kernel(self):

        return system.proc_kernel(self.__pid)

    def __get_proc_list(self):

        return system.proc_list(self.__which, self.__pid)

    def __get_proc_map(self):

        return system.proc_map(self.__pid)

    def __get_proc_mem(self):

        return system.proc_mem(self.__pid)

    def __get_proc_segment(self):

        return system.proc_segment(self.__pid)

    def __get_proc_signal(self):

        return system.proc_signal(self.__pid)

    def __get_proc_state(self):

        return system.proc_state(self.__pid)

    def __get_proc_time(self):

        return system.proc_time(self.__pid)

    def __get_proc_uid(self):

        return system.proc_uid(self.__pid)



    allfs            = property(__get_allfs, __set_allfs,
                                doc = "Getter/Setter for boolean allfs")
    iface            = property(__get_dev, __set_dev,
                                doc = "Getter/Setter for network interface")
    mountdir         = property(__get_mdir, __set_mdir,
                                doc = "Getter/Setter for mountdir")
    pid              = property(__get_pid, __set_pid,
                                doc = "Getter/Setter for pid")
    which            = property(__get_which, __set_which,
                                doc = "Getter/Setter for which")

    cpu              = property(FakeSelf(system.cpu), doc = "General CPU info")
    cpu_bogomips     = property(FakeSelf(system.cpu_bogomips),
                                         doc = "CPU bogomips")
    cpu_cache        = property(FakeSelf(system.cpu_cache),
                                doc = "CPU 2nd level cache")
    cpu_clock        = property(FakeSelf(system.cpu_speed), doc = "CPU clock")
    cpu_load         = property(FakeSelf(system.cpu_load), doc = "CPU load")
    cpu_model        = property(FakeSelf(system.cpu_model), doc =  "CPU model")

    fsusage          = property(__get_fsusage,
                                doc = "Information of given mounted filesystem")
    hostname         = property(FakeSelf(system.hostname),
                                doc = "The machine's hostname")
    kernel_version   = property(FakeSelf(system.kernel_version),
                                doc = "Kernel version")
    loadavg          = property(FakeSelf(system.loadavg),
                                doc = "Average load info")
    memory           = property(FakeSelf(system.mem), doc = "Memory information")
    mountlist        = property(FakeSelf(system.mountlist),
                                doc = "(All) mountpoints")
    net_devices      = property(FakeSelf(system.net_devices),
                                doc = "Available network devs")
    net_load         = property(__get_net_load, doc = "Netload of given device")
    net_speed        = property(__get_net_speed,
                                doc = "Incoming / Outgoing traffic speed of "
                                      "given device")
    net_state        = property(__get_net_state, doc = "Device up or down")

    operating_system = property(FakeSelf(system.operating_system),
                                doc = "Type of operating system")
    ppp              = property(__get_ppp, doc = "Point-to-Point information")
    proc_args        = property(__get_proc_args,
                                doc = "Command line args of process ID")
    proc_kernel      = property(__get_proc_kernel, doc = "")
    proc_list        = property(__get_proc_list, doc = "")
    proc_map         = property(__get_proc_map, doc = "")
    proc_mem         = property(__get_proc_mem, doc = "")
    proc_segment     = property(__get_proc_segment, doc = "")
    proc_signal      = property(__get_proc_signal, doc = "")
    proc_state       = property(__get_proc_state, doc = "")
    proc_time        = property(__get_proc_time, doc = "")
    proc_uid         = property(__get_proc_uid, doc = "")
    swap             = property(FakeSelf(system.swap),
                                doc = "Swap space information")
    uptime           = property(FakeSelf(system.uptime),
                                doc = "Uptime information")
    users            = property(FakeSelf(system.users), doc ="Number of users")



def get_class(): return System



import os
import sys
import traceback
import vfs


#
# Class for formatting errors. Hilights the bad line of code.
#
class _ErrorFormatter(object):

    __slots__ = ()

    #
    # Analyzes the given traceback and returns a developer-friendly error
    # output with markups.
    # If code is given, that string of code will be used as the source code,
    # instead of loading it.
    # If lineno is given, that number will be used as the linenumber. Only makes
    # sense together when code is given.
    # If deep_trace is True, the traceback history will be traversed as well.
    #
    def format(self, tb, code = None, lineno = -1, filename = "",
               deep_trace = False):

        # tracebacks are dangerous objects; to avoid circular references,
        # we have to drop references to a traceback ASAP

        # get recent traceback
        exc_type = tb[0]
        exc_value = tb[1]
        exc_tb = tb[2]

        # get filename and lineno
        if (not filename and lineno == -1):
            if (hasattr(exc_value, "filename") and
                  hasattr(exc_value, "lineno")):
                filename = exc_value.filename
                lineno = exc_value.lineno
            else:
                tbs = traceback.extract_tb(exc_tb)
                filename = tbs[-1][0]
                lineno = tbs[-1][1]
                del tbs            
        

        # print error message
        out = "\n"
        #out += "[EXC]%s:\n" % str(exc_type)
        out += "[EXC]%s\n" % str(exc_value)

        # dig into the traceback for additional information
        if (deep_trace):
            for trace in traceback.extract_tb(exc_tb):
                cntxt = trace[0]
                lno = trace[1]
                funcname = trace[2]
                out += "in %s: line %d %s\n" % (cntxt, lno, funcname)

        # load code from file if no code was specified
        if (not code):
            # get last traceback (otherwise we would load the wrong file for
            # hilighting)
            this_tb = exc_tb
            while (this_tb.tb_next):
                tmp = this_tb
                del this_tb
                this_tb = tmp.tb_next
                del tmp

            # get the .py file; we don't want .pyc or .pyo!
            path = this_tb.tb_frame.f_globals.get("__file__")
            del this_tb
            if (path and path[-4:-1] == ".py"): path = path[:-1]

            if (path and vfs.exists(path)):
                code = vfs.read_entire_file(path)
                filename = path
        #end if

        del exc_tb

        # find and hilight the bad line of code, while adding handy line numbers
        if (code):
            lines = code.splitlines()
            lno = 1
            for i in range(len(lines)):
                if (lno == lineno):
                    lines[i] = "[ERR]>%4d " % lno + lines[i]
                else:
                    lines[i] = "[---] %4d " % lno + lines[i]
                lno += 1
            #end for

            # take a small chop out of the code
            begin = max(0, lineno - 6)
            part = lines[begin:begin + 12]

            out += "[EXC]%s\n\n" % filename
            out += "\n".join(part)

        else:
            out += "[EXC]%s\n\n" % filename
            out += "[EXC]>> could not load source code for hilighting <<"

        #end if

        return out






#
# We are unable to load the source code if only a relative filename was
# available. Therefore, we have to extend the import handler in order to always
# give us an absolute path.
#
_old_imp = __import__
def _new_imp(name, globs = {}, locls = {}, fromlist = []):

    module = _old_imp(name, globs, locls, fromlist)
    # builtin modules have no "__file__" attribute, so we have to check for it
    if (hasattr(module, "__file__")):
        module.__file__ = os.path.abspath(module.__file__)
    return module

import __builtin__
__builtin__.__import__ = _new_imp



_singleton = _ErrorFormatter()
def ErrorFormatter(): return _singleton



from ConfigBoolean import ConfigBoolean
from ConfigButton import ConfigButton
from ConfigColor import ConfigColor
from ConfigEnum import ConfigEnum
from ConfigFloat import ConfigFloat
from ConfigFont import ConfigFont
from ConfigInteger import ConfigInteger
from ConfigString import ConfigString
from ConfigTitle import ConfigTitle
from ConfigUnit import ConfigUnit
from ConfigURI import ConfigURI

from ConfigDPI import ConfigDPI

from utils.HIGDialog import HIGDialog

import gtk
import gobject



class ConfigDialog(HIGDialog):
    """
      Class for the configuration dialog. This class handles the visualization
      of the dialog and loading / saving the configuration.
    """

    # mapping between item types and their widgets
    __ITEM_TABLE = {"boolean": ConfigBoolean,
                    "button": ConfigButton,
                    "color": ConfigColor,
                    "enum": ConfigEnum,
                    "float": ConfigFloat,
                    "font": ConfigFont,
                    "integer": ConfigInteger,
                    "string": ConfigString,
                    "title": ConfigTitle,
                    "unit": ConfigUnit,
                    "uri": ConfigURI,
                    "dpi": ConfigDPI
                    }


    def __init__(self):

        # the instantiated items
        self.__children = []

        # getter, setter, and caller for getting and settings values and
        # calling callbacks
        self.__getter = None
        self.__setter = None
        self.__caller = None

        # a banner
        self.__banner = None


        HIGDialog.__init__(self, buttons=(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE),
                           self_destroy = False)

        self.__tooltips = gtk.Tooltips()
        self.__tooltips.enable()


        def destroy(*args):
            self.hide()
            return True  # return True so that the window doesn't get destroyed

        self.connect("response", destroy)
        self.connect("delete-event", destroy)  # catch the close button as well


    def _set_setter(self, setter): self.__setter = setter
    def _set_getter(self, getter): self.__getter = getter
    def _set_caller(self, caller): self.__caller = caller


    def _set_value(self, key, value, datatype):

        self.__setter(key, value, datatype)


    def _get_value(self, key):

        return self.__getter(key)


    def _invoke_callback(self, name, *args):

        try:
            self.__caller(name, *args)
        except KeyError:
            raise UserError(_("Invalid Function Call"),
                           _("There is no function called <b>%s</b>.\nThis "
                             "means that there's an error in the desklet's "
                             "code. You should inform the author of the "
                             "desklet about this problem." % (name,)))

        except StandardError, exc:
            log("An unknown error occured: %s" % (exc,))



    #
    # Creates a banner. This method has to be called before building the dialog
    # if a banner is desired.
    #
    def set_banner(self, icon, text):

        self.__banner = gtk.Frame()
        self.__banner.set_shadow_type(gtk.SHADOW_IN)
        self.__banner.show()

        b_box = gtk.HBox(spacing = 12)
        b_box.set_border_width(6)
        b_box.show()
        self.__banner.add(b_box)

        if (icon):
            try:
                # for him who loves small try blocks: you don't gain anything
                # from splitting up this block except for badly readable code ;)
                from utils import vfs
                data = vfs.read_entire_file(icon)
                loader = gtk.gdk.PixbufLoader()
                loader.write(data, len(data))
                loader.close()
                pbuf = loader.get_pixbuf()

                # scale icon down while preserving aspect ratio
                width = pbuf.get_width()
                height = pbuf.get_height()
                scale = 48 / float(height)
                width *= scale
                if (abs(scale - 1.0) > 0.001):
                    pbuf = pbuf.scale_simple(int(width), 48, 3)

                b_icon = gtk.Image()
                b_icon.set_from_pixbuf(pbuf)
                b_icon.show()
                b_box.pack_start(b_icon, False, False)

            except:
                pass

        #end if

        b_label = gtk.Label(text)
        b_label.set_use_markup(True)
        b_label.show()
        b_box.pack_start(b_label, True, True)



    #
    # Creates a configuration item of the given type.
    #
    def __create_config_item(self, itype, settings):

        try:
            itemclass = self.__ITEM_TABLE[itype]
        except KeyError:
            log("No such preferences item: %s" % (itype,))
            return None

        item = itemclass(itype, self._get_value, self._set_value,
                         self._invoke_callback)

        return item



    #
    # Adds a line of configuration widgets to the given page.
    #
    def __add_line(self, page, page_lines, indent, widgets, help):

        ebox = gtk.EventBox()
        ebox.show()

        if (len(widgets) == 2):
            page.attach(widgets[0], 0, 1, page_lines - 1, page_lines,
                        gtk.FILL, 0, indent, 3)
            ebox.add(widgets[1])
            page.attach(ebox, 1, 2, page_lines - 1, page_lines,
                        gtk.EXPAND | gtk.FILL, 0, 0, 3)

        else:
            ebox.add(widgets[0])
            page.attach(ebox, 0, 2, page_lines - 1, page_lines,
                        gtk.EXPAND | gtk.FILL, 0, indent, 3)

        if (help):
            self.__tooltips.set_tip(ebox, help)



    #
    # Builds up the dialog. This must be done before loading the configuration.
    #
    def build(self, items):

        if (self.__banner):
            self.vbox.pack_start(self.__banner, False, False)

        if (not items):
            label = gtk.Label(_("This desklet is not configurable."))
            label.show()
            self.vbox.pack_start(label, False, False)

        page_lines = 0
        page = gtk.Table(1, 2)
        page.show()
        notebook = None
        item_list = []
        self.__children = []

        for itype, settings in items:
            if (itype == "page"):
                if (not notebook):
                    align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
                    align.show()
                    notebook = gtk.Notebook()
                    notebook.set_property("border-width", 6)
                    notebook.show()
                    align.add(notebook)
                    self.vbox.pack_start(align, False, False, 0)

                label = settings.get("label", "")
                tab = gtk.Label(label)
                tab.show()
                box = gtk.VBox()
                box.set_property("border-width", 12)
                box.show()
                box.add(page)
                notebook.append_page(box, tab)

                # create a new page
                page_lines = 0
                page = gtk.Table(1, 2)
                page.show()

            elif (itype == "item"):
                label = settings.get("label", "")
                value = settings.get("value", "")
                item_list.append((label, value))

            else:
                configitem = self.__create_config_item(itype, settings)
                if (not configitem): continue

                widgets = configitem.get_widgets()
                if (itype not in ("title", "button")):
                    configitem.set_prop_from_string("bind", settings["bind"])
                if (itype == "enum"): configitem.set_prop("items", item_list)
                for k, v in settings.items():
                    configitem.set_prop_from_string(k, v)

                help = settings.get("help", "")

                indent = (itype != "title") and 12 or 0
                page_lines += 1
                page.resize(page_lines, 2)
                if (itype == "title" and page_lines != 1):
                    # keep some HIGgy space above the title, but not if it's
                    # the first line
                    w, h = widgets[0].size_request()
                    widgets[0].set_size_request(-1, h + 12)
                self.__add_line(page, page_lines, indent, widgets, help)
                if (itype != "title"):
                    self.__children.append(configitem)

                item_list = []

            #end if

        #end for

        # add the only page manually if there's no notebook
        if (not notebook):
            self.vbox.pack_start(page, False, False, 0)


    #
    # Returns a list of the config items.
    #
    def get_config_items(self):

        return self.__children


    #
    # Sets the current working directory.
    #
    def set_path(self, path):

        for c in self.get_config_items(): c.set_path(path)


    #
    # Refine the show() method to update the configuration widgets.
    #
    def show(self):

        # have the children update themselves
        for c in self.get_config_items(): c.update()

        self.present()

        # run the updater
        gobject.timeout_add(500, self.__updater)



    def __updater(self):

        for c in self.get_config_items(): c.update()

        # stop the timer if the dialog is no longer visible
        if (self.get_property("visible")):
            return True
        else:
            return False




"""
This module implements a simple tuple transmission protocol. It's kept simple
to be implementable in other languages as well.

The protocol can only transmit lists of strings, which it splits up into chunks.
A chunk consists of a LENGTH byte, the actual PAYLOAD, and a STATUS byte.

A single string is split up into several chunks if it exceeds the maximum length
which can be specified by the LENGTH byte (255), otherwise a string is one
chunk.

The STATUS byte after every chunk tells if the chunk is

 - continued in the next chunk (CONT)
 - the last chunk of a string (NEXT)
 - the last chunk in the transmission (END)

In order to handle empty lists without getting too complicated, all lists are
extended by an arbitrary first element which just gets ignored.


Example:

  ["Hello", "World!"] is transmitted as:

  01  00 01   05   48 65 6C 6C 6F 01   06  57 6F 72 6C 64 21 02

  (1) ?  NEXT (5)  H  e  l  l  o  NEXT (6) W  o  r  l  d  !  END
"""


class XDRError(RuntimeError):
    pass


_SEND_ERROR = "--SEND ERROR--"


_CONT = chr(0)
_NEXT = chr(1)
_END = chr(2)



def send(s, *args):

    args = ["\0"] + list(args)
    while (args):
        a = args.pop(0)
        
        chunks = [ a[i:i + 0xff] for i in range(0, len(a), 0xff) ]
        while (chunks):
            c = chunks.pop(0)
            s.send(chr(len(c)))
            s.send(c)
            if (chunks): s.send(_CONT)

        if (args): s.send(_NEXT)
        
    s.send(_END)

    
def send_error(s):

    send(s, _SEND_ERROR)


def recv(s):

    args = []
    chunk = ""
    while (True):
        try:
            length = ord(s.recv(1))
        except:
            raise XDRError

        if (length): chunk += s.recv(length)

        flag = s.recv(1)
        if (flag == _CONT): continue

        args.append(chunk)
        chunk = ""
        
        if (flag == _END): break

    #end while

    return args[1:]



#
# Layout algorithms.
#
# A layout algorithm is a function f: (step, args) -> (x, y, relative)
# A step value of n means that the coordinates of the nth child are being
# requested.
#

class LayoutError(StandardError):
    pass


class Layouter:

    __slots__ = ('__layout',)
    __LAYOUT_TYPES = ('horizontal', 'vertical', 'grid')

    def __init__(self, layout):

        self.__layout = layout

        self.__validate_layout()



    def __validate_layout(self):

        if (not self.__layout in self.__LAYOUT_TYPES):
            log("Layout type %s isn't supported." % (self.__layout,))
            raise LayoutError



    def layout(self, step, args):

        if (self.__layout == "horizontal"):
            if (not args):
                return ("x", 0, 0)
            else:
                delta = int(args[0])
                x = delta * step
                y = 0

                return (None, x, y)

        elif (self.__layout == "vertical"):
            if (not args):
                return ("y", 0, 0)
            else:
                delta = int(args[0])
                y = delta * step
                x = 0

                return (None, x, y)

        else:
            width, deltax, deltay = args
            x = (step % int(width)) * int(deltax)
            y = (step / int(width)) * int(deltay)

            return (None, x, y)




from BookmarkList import BookmarkList
from ListView import ListView
from SearchBar import SearchBar
from utils.Observable import Observable


try:
    set
except NameError:
    from sets import Set as set
    
import gtk


#
# Class for item browser widgets.
#
class ItemBrowser(Observable):

    OBS_ACTIVATE_ITEM = 0


    def __init__(self, search_handler, item_renderer, *columns):

        self.__search_handler = search_handler

        # build bookmarks list
        self.__bookmarks = BookmarkList()
        self.__browser = gtk.VBox()

        # build search bar
        self.__searchbar = SearchBar()
        self.__searchbar.show_all()
        self.__searchbar.add_observer(self.__on_search)
        self.__browser.pack_start(self.__searchbar, False, False, 0)

        # build list view
        scrwin = gtk.ScrolledWindow()
        scrwin.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.__browser.pack_end(scrwin, True, True, 0)

        self.__list = ListView(columns, item_renderer)
        self.__list.connect("row-activated", self.__on_activate_item)
        scrwin.add(self.__list)


    #
    # Returns the widgets.
    #
    def get_widgets(self):

        return (self.__bookmarks, self.__browser)


    #
    # Adds a category to the bookmarks list.
    #
    def add_category(self, name, label, items):

        self.__bookmarks.add_category(name, name)

        items = list(set(items))
        items.sort(lambda a,b: cmp(a[1], b[1]))
        for icon, label, expr in items:
            if (icon):
                pixbuf = gtk.gdk.pixbuf_new_from_file(icon)
            else:
                pixbuf = None
            
            self.__bookmarks.add_item(name, pixbuf, label,
                                      self.__on_select_entry, expr)
        #end for
        
        self.__bookmarks.select_category(0) 


    #
    # Sets a search query manually.
    #
    def set_query(self, expr):
        
        self.__searchbar.set_query(expr)


    #
    # Returns the currently selected item.
    #
    def get_selected_item(self):

        return self.__list.get_selected_item()
    

    def __on_select_entry(self, expr):

        self.set_query(expr)


    def __on_search(self, src, cmd, expr):
        assert (self.__search_handler)
        
        self.__list.clear()
        matches = self.__search_handler(expr)

        if (matches):
            for item in matches:
                self.__list.add_item(item)

        return matches


    def __on_activate_item(self, *args):

        item = self.__list.get_selected_item()
        self.update_observer(self.OBS_ACTIVATE_ITEM, item)



from TargetGroup import TargetGroup
from utils.datatypes import *
from layout import Unit

import gtk


#
# The gauge is a decorator that can turn anything into a gauge.
#
class TargetGauge(TargetGroup):

    __HORIZONTAL = "horizontal"
    __VERTICAL = "vertical"


    def __init__(self, name, parent):

        TargetGroup.__init__(self, name, parent)

        self._register_property("fill", TYPE_FLOAT,
                                self._setp_fill, self._getp)
        self._register_property("orientation", TYPE_STRING,
                                self._setp_orientation, self._getp)
        self._setp("fill", 100)
        self._setp("orientation", self.__HORIZONTAL)


    def new_child(self, childtype, settings, children):

        child = TargetGroup.new_child(self, childtype, settings, children)
        child.add_observer(self.__on_observe_child)


    def __on_observe_child(self, src, cmd, *data):

        if (cmd == src.OBS_GEOMETRY):
            self.set_prop("fill", self.get_prop("fill"))


    #
    # "fill" property.
    #
    def _setp_fill(self, key, value):

        value = max(0, min(value, 100))
        x, y, width, height = self._get_child().get_geometry()
        width = width.as_pt()
        height = height.as_pt()
        orientation = self.get_prop("orientation")
        if (orientation == self.__HORIZONTAL):
            width = int(width * (value / 100.0))
            self.set_prop("width", Unit.Unit(width, Unit.UNIT_PT))
        else:
            height = int(height * (value / 100.0))
            self.set_prop("height", Unit.Unit(height, Unit.UNIT_PT))

        self._setp(key, value)


    #
    # "orientation" property.
    #
    def _setp_orientation(self, key, value):

        self.set_prop("fill", self.get_prop("fill"))
        self._setp(key, value)



from main import USERHOME
from datatypes import *
from layout import Unit
import pwstore


def str2type(dtype, value):

    if (dtype == TYPE_STRING):
        return value

    elif (dtype == TYPE_SECRET_STRING):
        return pwstore.retrieve(USERHOME, value)

    elif (dtype == TYPE_BOOL):
        return (value in ("True", "true", "1", "1L"))

    elif (dtype == TYPE_INT):
        return int(value)

    elif (dtype == TYPE_FLOAT):
        return float(value)

    elif (dtype == TYPE_LIST):
        v = value.replace("\\,", "@@COMMA@@")
        parts = v.split(",")
        return [ p.strip().replace("@@COMMA@@", ",") for p in parts ]

    elif (dtype == TYPE_UNIT):
        return Unit.Unit(string = value)

    elif (dtype == TYPE_UNIT_LIST):
        parts = value.split(",")
        return [ str2type(TYPE_UNIT, unit.strip()) for unit in parts ]

    else:
        return value


#
# This method is only used by Sensor and is thus deprecated.
#
def type2str(dtype, value):

    if (dtype == TYPE_LIST and isinstance(value, list)):
        return ",".join(value)

    elif (dtype == TYPE_SECRET_STRING):
        return pwstore.store(USERHOME, value)

    elif (dtype == TYPE_BOOL):
        return value and "true" or "false"

    elif (dtype == TYPE_UNIT):
        val, unit = value
        if (unit != UNIT_PX): return str(val) + unit
        else: return str(val)

    else:
        return str(value)



from libdesklets.controls import Control
from IArrayBuffer import IArrayBuffer


#----------------------------------------------------------------------------------#
#
# ArrayBuffer Class
#
#
class ArrayBuffer(Control, IArrayBuffer):

        #
        # ArrayBuffer Constructor
        #
        def __init__(self):

                self.__buffer           = []
                self.__size             = 0
                self.__cursor           = 0
                self.__window_pos       = 0
                self.__window_size      = 0
                self.__filltype         = None

                Control.__init__(self)


        #
        # Get the current object at cursor position
        #
        def __read(self):

                return self.__buffer[self.__window_pos:self.__window_pos + \
                                     self.__window_size]


        #
        # Get the entire buffer
        #
        def __read_all(self):

                return self.__buffer


        #
        # Deletes a single line from the buffer
        #
        def __delete(self, pos):

                self.__buffer = [self.__filltype] + self.__buffer[:pos] + \
                                 self.__buffer[pos + 1:]


        #
        # Fill the buffer with an object
        #
        def __fill(self, obj):

                for i in range(self.__size):
                        self.__buffer[i] = obj


        #
        # Write object to current cursor position within buffer
        #
        def __write(self, objArr):

                n = len(objArr)

                if self.__cursor + n > self.__size:
                        if n >= self.__size:
                                tmp = n - self.__size
                                self.__buffer = objArr[tmp:]

                        else:
                                tmp = self.__size - n
                                self.__buffer = self.__buffer[self.__cursor - \
                                                              tmp:self.__cursor]
                                self.__buffer += objArr

                        self.__cursor = self.__size

                else:
                        self.__buffer[self.__cursor:] = objArr + \
                                                        self.__buffer[self.__cursor + n:]
                        self.__cursor += n


        #
        # Get the current viewable window position
        #
        def __get_window_pos(self):

                return self.__window_pos


        #
        # Get the filltype
        #
        def __get_filltype(self):

                return self.__filltype


        #
        # Sets the filltype used when resizing the buffer
        #
        def __set_filltype(self, obj):

                self.__filltype = obj


        #
        # Set the viewable window position
        #
        def __set_window_pos(self, pos):

                if pos < 0:
                        pos = 0

                if pos + self.__window_size > self.__size:
                        pos = self.__size - self.__window_size

                self.__window_pos = pos


        #
        # Get the size of the viewable window
        #
        def __get_window_size(self):

                return self.__window_size


        #
        # Set the size of the viewable window
        #
        def __set_window_size(self, size):

                if size > self.__size:
                        size = self.__size

                self.__window_size = size


        #
        # Get the size of the buffer
        #
        def __get_size(self):

                return self.__size


        #
        # Set the size of the buffer
        #
        def __set_size(self, size):

                if size > self.__size:

                        tmp = [ self.__filltype for i in range(size - self.__size) ]

                        self.__buffer = tmp + self.__buffer

                else:
                        self.__buffer = self.__buffer[self.__size - size:]

                self.__size = size


        #
        # Get the current buffer cursor position
        #
        def __get_cursor(self):

                return self.__cursor


        #
        # Set the position of the buffer cursor
        #
        def __set_cursor(self, pos):

                if pos < 0:
                        pos = 0

                if pos > self.__size:
                        pos = self.__size

                self.__cursor = pos


        #
        # Interface
        #
        read        = property(fget = __read,
                               doc = "Return the viewable window array")
        read_all    = property(fget = __read_all,
                               doc = "Return the entire buffer array")
        delete      = property(fset = __delete,
                               doc = "Delete line from buffer")
        write       = property(fset = __write,
                               doc = "Write line at position cursor")
        fill        = property(fset = __fill,
                               doc = "Fill the buffer with an object")
        filltype    = property(fget = __get_filltype, fset = __set_filltype,
                               doc = "Get/Set the empty space fill type")
        window_pos  = property(fget = __get_window_pos, fset = __set_window_pos,
                               doc = "Get/Set viewable window position")
        window_size = property(fget = __get_window_size, fset = __set_window_size,
                               doc = "Get/Set viewable window size")
        size        = property(fget = __get_size, fset = __set_size,
                                doc = "Get/Set buffer size")
        cursor      = property(fget = __get_cursor, fset = __set_cursor,
                               doc = "Get/Set cursor position")


#----------------------------------------------------------------------------------#

def get_class(): return ArrayBuffer




from plugin.PluginRegistry import PluginRegistry
from main import HOME, USERHOME

import os


#
# Registry for shell plugins.
#
class _PluginRegistry:

    def __init__(self):

        # prevent from recursion
        self.__cache = {}
        
        # table: query -> result
        self.__query_cache = {}

        path = os.path.dirname(__file__)
        self.__registry = PluginRegistry(os.path.join(path, "plugins"))



    def get_plugins_by_pattern(self, key, pattern):

        query = "(MATCH '%s' '%s')" % (key, pattern)
        if (not query in self.__query_cache):
            result = self.__registry.get_plugins(query)
            self.__query_cache[query] = result[:]
        else:
            result = self.__query_cache[query]

        ret = []

        for c in result:

            if c not in self.__cache:
                self.__cache[c] = c()
                self.__cache[c].init()

            ret.append( self.__cache[c] )

        return ret


    #
    # Returns a control providing the given interfaces.
    #
    def get_plugin(self, name, *interfaces):

        result = self.get_plugins_by_pattern("name", name)
        if (result):
            return result[0]


# singleton object
registry = _PluginRegistry()



from main import HOME, SENSORPATHS

import os, sys

if "." not in sys.path: sys.path.append(".")
if HOME not in sys.path: sys.path.append(HOME)


#
# Factory class for dynamically loaded Sensors.
#
class _SensorFactory:

    def __init__(self):

        # a set to remember the modules loaded so far
        self.__modules = {}



    #
    # Creates and returns a Sensor of the given type or None if the Sensor
    # could not be created.
    #
    def create_sensor(self, name, args):

        # find and import the sensor
        oldcwd = os.getcwd()
        sensor = None
        module = None

        try:

            for p in SENSORPATHS:

                try:
                    os.chdir(p)
                except OSError:
                    continue

                # reload old modules to force using the latest version; this
                # means we can edit modules while running gDesklets
                # TODO: clean up

                try:
                    sensordir = os.path.abspath(name)
                    if (not os.path.exists(sensordir)): continue

                    if (name in sys.modules): del sys.modules[name]
                    module = __import__(name)
                    path = sensordir

                    break

                except ImportError, e:
                    from utils.ErrorFormatter import ErrorFormatter
                    details = ErrorFormatter().format(sys.exc_info())
                    print details
                    continue


            if (module):
                self.__modules[name] = (module, path)

                # initalize the sensor
                os.chdir(path)
                sensor = module.new_sensor(args)

            else:
                print "Could not load sensor %(name)s" % vars()
                from utils.ErrorFormatter import ErrorFormatter
                details = ErrorFormatter().format(sys.exc_info())
                print details

                raise UserError(_("Could not find sensor '%(name)s'") % vars(),
                      _("A sensor could not be found. This usually means that "
                        "it has not been installed."),
                                show_details = False)


            return sensor

        finally:
            os.chdir(oldcwd)


_singleton = _SensorFactory()
def SensorFactory(): return _singleton



from main import _
from utils.HIGDialog import HIGDialog

import gtk


#
# Class for the display configurator. It displays the configurators of the
# sensors of a display.
#
class DisplayConfigurator(HIGDialog):

    def __init__(self, sensorconfigurators):

        HIGDialog.__init__(self, buttons=(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))
        self.set_property("title", _("Configuration"))

        def destroy(*args): self.destroy()

        self.connect("response", destroy)

        # close functions
        pages = []
        for c in sensorconfigurators:
            if (c):
                lbl = gtk.Label(c.get_name())
                lbl.show()
                pages.append((c, lbl))

        # use a special page when there are no config options
        if (not pages):
            lbl = gtk.Label(_("This desklet is not configurable."))
            lbl.show()
            pages.append((lbl, None))

        # only use the notebook when there are more than one pages
        if (len(pages) == 1):
            self.vbox.add(pages[0][0])
        else:
            align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
            align.show()
            notebook = gtk.Notebook()
            notebook.set_property("border-width", 6)
            notebook.show()
            align.add(notebook)
            self.vbox.pack_start(align, False, False, 0)
            for page, tab in pages:
                notebook.append_page(page, tab)

        self.show()




#
# Abstraction layer for VFS operations to move gnomevfs dependency out of the
# core.
#

USE_GNOMEVFS = True

import urllib

try:
    import gnomevfs

    OPEN_READ = gnomevfs.OPEN_READ
    OPEN_WRITE = gnomevfs.OPEN_WRITE

except ImportError:
    try:
        import gnome.vfs as gnomevfs

        OPEN_READ = gnomevfs.OPEN_READ
        OPEN_WRITE = gnomevfs.OPEN_WRITE

    except ImportError:
        OPEN_READ = "r"
        OPEN_WRITE = "w"

        USE_GNOMEVFS = False



def escape_path(uri):

    try:
        return gnomevfs.escape_string(uri)
    except:
        return escape_path_urllib(uri)

def escape_path_urllib(uri):

    return urllib.quote(uri)



def unescape_path(uri):

    return urllib.unquote(uri)



#
# Reads the entire file and returns its contents.
#
def read_entire_file(uri):

    try:
        uri = gnomevfs.read_entire_file(uri)
    except:
        log("Warning: Couldn't read file \"%s\"." % (uri,))
        raise

    return uri

def read_entire_file_urllib(uri):

    if (not "://" in uri): uri = "file://" + uri

    try:
        uri = urllib.urlopen(uri).read()
    except:
        log("Warning: Couldn't read file \"%s\"." % (uri,))
        raise

    return uri



#
# Opens the given URI and returns a file descriptor.
#
def open(uri, mode = OPEN_READ):

    try:
        fd = gnomevfs.open(uri, mode)
    except:
        log("Warning: Couldn't open file \"%s\"." % (uri,))
        return

    return fd

def open_urllib(uri, mode = OPEN_READ):

    if (not "://" in uri): uri = "file://" + uri

    try:
        fd = urllib.urlopen(uri)
    except:
        log("Warning: Couldn't open file \"%s\"." % (uri,))
        return

    return fd



#
# Returns whether the given URI exists.
#
def exists(uri):

    return gnomevfs.exists(gnomevfs.URI(uri))

def exists_urllib(uri):

    if (not "://" in uri): uri = "file://" + uri

    try:
        urllib.urlopen(uri)
    except:
        return False
    else:
        return True



if (not USE_GNOMEVFS):
    read_entire_file = read_entire_file_urllib
    open = open_urllib
    exists = exists_urllib
    escape_path = escape_path_urllib



from libdesklets.controls import Control

from ICalendar import ICalendar

import calendar
import time
import datetime

class Calendar(Control, ICalendar):

    # Monday, 13th of December 2004
    __DAYS = tuple([ time.strftime('%A', datetime.date(2004, 12, 13 + d).timetuple())
                     for d in range(0, 7) ])



    __MONTHS = tuple([ time.strftime('%B', datetime.date(2004, m, 1).timetuple())
                       for m in range(1, 13) ])



    def __init__(self):
        self.__time = Null
        Control.__init__(self)


    def __get_days(self):
        return self.__DAYS


    def __get_months(self):
        return self.__MONTHS


    def __set_time(self, time):

        assert 'ITime:9y703dqtfnv4w373caserz68r' in time.get_interfaces_id()
        self.__time = time


    def __ymd(self):
        return self.__time.date
    

    def __get_day(self):
        wd = calendar.weekday( *self.__ymd() )
        return self.__DAYS[ wd ]


    def __get_month(self):
        return self.__MONTHS[ self.__ymd()[1] - 1 ]



    day    = property(fget = __get_day)
    days   = property(fget = __get_days)
    month  = property(fget = __get_month)
    months = property(fget = __get_months)
    time   = property(fset = __set_time)


def get_class(): return Calendar



from Generic import Generic

import re


class PPC(Generic):

    def __init__(self):

        Generic.__init__(self)

        def _get_model():
            r = re.compile('^cpu\s+:\s+(.+)$', re.M)
            m = r.search( self._read_cpuinfo() )
            return m.group(1)

        def _get_cache():
            r = re.compile('^L2 cache\s+:\s+(\d+)', re.M)
            m = r.search( self._read_cpuinfo() )
            return int(m.group(1))

        # CPU model and cache size never changes
        self.__model_name = _get_model()
        self.__cache_size = _get_cache()

        # the cpu speed might change (laptops have mobile CPUs)
        self.__speed      = re.compile('^clock\s+:\s+(\d+)', re.M)



    def cpu_speed(self):
        """
        @return : current clock of installed processor
        @rtype  : float
        """

        m = self.__speed.search( self._read_cpuinfo() )
        return float(m.group(1))



    def cpu_cache(self):
        """
        @return : 2nd level cache of installed processor
        @rtype  : int
        """

        return self.__cache_size



    def cpu_model(self):
        """
        @return : model/type of installed processor
        @rtype  : str
        """

        return self.__model_name



from utils.Trie import Trie
from utils.QueryParser import QueryParser
from plugin.Interface import Interface
from main import HOME

import os
import sys


if "." not in sys.path: sys.path.append(".")
if HOME not in sys.path: sys.path.append(HOME)


#
# Registry class for finding and indexing plugins.
#
class PluginRegistry:

    # the keys for indexing plugins
    INTERFACE = "interface"
    TAZINTERFACE = "tazinterface"
    NAME = "name"
    MTIME = "mtime"


    def __init__(self, *repositories):

        # the trie for indexing the registered plugins
        self.__trie = Trie()

        # the query parser that is wrapped around the trie
        self.__queryparser = QueryParser(self.__trie)

        for repo in repositories:
            self.__scan_repository(repo)



    #
    # Scans the given repository for plugins and registers them.
    #
    def __scan_repository(self, repo):

        if (os.path.exists(repo)): files = os.listdir(repo)
        else: return

        for f in files:
            pluginpath = os.path.join(repo, f)
            if (os.path.exists(os.path.join(pluginpath, "__init__.py"))):
                cwd = os.getcwd()

                try:
                    os.chdir(repo)
                except OSError, exc:
                    print >> sys.stderr, "Couldn't chdir to %s" % (repo,)

                try:
                    module = __import__(f)
                    self.__register_plugin(module)
                except StandardError, exc:
                    print >> sys.stderr, "%s\n%s in %s is NOT a valid plugin!" \
                                         % (exc, f, pluginpath)

                os.chdir(cwd)

            elif (os.path.isdir(pluginpath)): self.__scan_repository(pluginpath)



    def __register_plugin(self, module):

        clss = module.get_class()
        path = os.path.abspath(os.path.dirname(module.__file__))
        clss._path = path
        interfaces = Interface.get_interfaces(clss)
        name = clss.__name__
        mtime = os.path.getmtime(module.__file__)

        self.__trie.insert([self.NAME] + list(name), clss)
        self.__trie.insert([self.MTIME] + list(`mtime`), clss)

        for iface in interfaces:
            ident = Interface.get_id(iface)
            taz_ident = Interface.get_taz_style_id(iface)
            self.__trie.insert([self.INTERFACE] + list(ident), clss)
            self.__trie.insert([self.TAZINTERFACE] + list(taz_ident), clss)


    def get_plugins(self, query):

        return self.__queryparser.parse(query)




import gtk
import gobject
import sys
import traceback

from cStringIO import StringIO
from HIGDialog import HIGDialog
from LogView import LogView


# define some Dialog icons
_ERROR = gtk.STOCK_DIALOG_ERROR
_INFO = gtk.STOCK_DIALOG_INFO
_QUESTION = gtk.STOCK_DIALOG_QUESTION
_WARNING = gtk.STOCK_DIALOG_WARNING

# we only want to display one dialog at a time, so let's queue them
_dialog_queue = []

# IDs which are to skip
_skip_ids = []

# remember the previous message to avoid displaying the same message twice
# in a sequence
_last_message = None


#
# Adds a details button to the given dialog.
#
def _set_details(dialog, details):

    vbox1 = gtk.VBox()
    vbox2 = dialog.vbox
    vbox2.pack_start(vbox1)
    align1 = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
    align1.set_property("border-width", 6)
    align1.show()
    vbox2.pack_start(align1)

    align2 = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
    align2.set_property("border-width", 6)
    align2.show()
    details = details.rstrip()
    expander = gtk.expander_new_with_mnemonic(
        _("_Details (%d lines)") % len(details.splitlines()))
    expander.show()
    viewport = gtk.ScrolledWindow()
    viewport.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)
    viewport.show()
    lbl = LogView()
    lbl.append(details)
    lbl.show()
    nil, height = lbl.get_size_request()
    width, nil = vbox2.get_size_request()
    viewport.set_size_request(width, min(height, 480))
    viewport.add_with_viewport(lbl)
    expander.add(viewport)
    align2.add(expander)

    vbox1.show()
    vbox1.pack_start(align2)


#
# Queues the given dialog for displaying.
#
def _queue_dialog(ident, dialog):

    def proceed(*args):
        if (not _dialog_queue): return
        _dialog_queue.pop(0)
        if (not _dialog_queue): return
        
        ident, dialog = _dialog_queue[0]

        if (not ident in _skip_ids):
            dialog.present()
        else:
            dialog.destroy()
            proceed()
            

    dialog.connect("destroy", proceed)

    # display the dialog immediately if there are no others in the queue
    _dialog_queue.append((ident, dialog))
    if (len(_dialog_queue) == 1):
        dialog.present()


#
# Removes all dialogs associated with the given ID from the queue.
#
def forget(ident_to_forget):

    q = []
    for ident, dialog in _dialog_queue:
        if (ident != ident_to_forget): q.append((ident, dialog))
    _dialog_queue[:] = q

    if (ident_to_forget in _skip_ids): _skip_ids.remove(ident_to_forget)


#
# Displays an error dialog. Errors are critical and the program terminates
# afterwards.
#
def error(primary, secondary):

    dialog = HIGDialog((gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE),
                       _ERROR, primary, secondary)
    gtk.threads_enter()
    dialog.run()
    gtk.threads_leave()
    sys.exit(1337)


#
# Displays an information dialog.
#
def info(primary, secondary):

    dialog = HIGDialog((gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE),
                       _INFO, primary, secondary)
    _queue_dialog(0, dialog)


#
# Displays a question dialog.
#
def question(icon, primary, secondary, *buttons):

    def responder(src, response):
        callback = buttons[response][1]
        if (callback): callback()

    response = 0
    btns = []
    for label, callback in buttons:
        btns.append(label)
        btns.append(response)
        response += 1

    dialog = HIGDialog(tuple(btns), _QUESTION, primary, secondary)
    dialog.connect("response", responder)
    dialog.show()


#
# Displays a warning dialog.
#
def warning(primary, secondary, details = "", force = False):

    global _last_message

    # don't show the same dialog twice in a sequence
    if (force): _last_message = ""
    if (_last_message == (primary, secondary, details)): return
    else: _last_message = (primary, secondary, details)

    dialog = HIGDialog((gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE),
                       _WARNING, primary, secondary)


    if (details):
        _set_details(dialog, details)

    _queue_dialog(0, dialog)


#
# Displays a user error dialog. This dialog is for hilighting invalid lines
# of code and is associated with a display instance.
#
def user_error(ident, primary, secondary, details = ""):

    if (ident in _skip_ids): return

    dialog = HIGDialog((gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE),
                       _WARNING, primary, secondary)


    if (details):
        _set_details(dialog, details)

    def f(src, ident):
        if (src.get_active() and not ident in _skip_ids):
            _skip_ids.append(ident)
        elif (not src.get_active() and ident in _skip_ids):
            _skip_ids.remove(ident)
        
    vbox = dialog.vbox
    chkbtn = gtk.CheckButton(_("_Ignore errors from this desklet"))
    chkbtn.connect("toggled", f, ident)
    chkbtn.show()
    vbox.pack_start(chkbtn)

    _queue_dialog(ident, dialog)

    


#
# Use the new filechoose if possible, or fallback to the old one
#
def fileselector(title, callback_ok, callback_cancel, *args):

    def handler(src, response):
        if (response == gtk.RESPONSE_OK):
            if (callback_ok): callback_ok(src, *args)
        else:
            if (callback_cancel): callback_cancel(src, *args)
            else: src.destroy()


    # do we have FileChooserDialog available?
    try:
        fsel = gtk.FileChooserDialog(title, None,
                                     gtk.FILE_CHOOSER_ACTION_OPEN,
                                     (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                      gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        fsel.connect("response", handler)

    # no, then use the old FileSelection
    except:
        def f(btn, fsel, response): handler(fsel, response)

        fsel = gtk.FileSelection()

        if (title): fsel.set_title(title)
        fsel.ok_button.connect("clicked", f, fsel, gtk.RESPONSE_OK)
        fsel.cancel_button.connect("clicked", f, fsel, gtk.RESPONSE_CANCEL)

    fsel.show()




from libdesklets.controls import Interface, Permission

class ICalendar(Interface):

    day = Permission.READ
    days = Permission.READ
    month = Permission.READ
    months = Permission.READ
    time = Permission.WRITE




from layout.LayoutObject import LayoutObject
from config.DisplayConfigger import DisplayConfigger
from config.StateSaver import DefaultStateSaver
from DisplayTarget import DisplayTarget
from TargetGroup import TargetGroup
from MenuItem import MenuItem
from main import UNSET_COORD
from scripting.Script import Script
from scripting.Scriptlet import Scriptlet
from utils import actionparser
from utils.Struct import Struct
from utils.Observable import Observable
from layout import Unit
from utils import vfs
from utils.tiling import Tiling
import utils

import utils.dialog

import gobject
import gtk
import time
import random
import sys
import os
import re


# The name of the namespace where all display element are accessible.
_ROOT = "Dsp"


#
# Class for display windows.
#
class Display(gtk.HBox, Observable):

    __slots__ = ('__sensor_controls', '__arrays',
                 '__group', '__script',
                 '__configurator', '__path', '__display_file', '__id',
                 '__last_pointer_pos', '__pointer_pos', '__is_sensitive',
                 '__cursor_stack')

    # observer commands
    OBS_CLOSE = 0
    OBS_RESTART = 1
    OBS_GEOMETRY = 2
    OBS_SHAPE = 3
    OBS_TITLE = 4
    OBS_ICON = 5
    OBS_FLAGS = 6
    OBS_BORDERS = 7
    OBS_CURSOR = 8
    OBS_CLOSED = 9


    # event types
    EVENT_MOTION = 0
    EVENT_PRESS = 1
    EVENT_RELEASE = 2
    EVENT_KEY_PRESS = 3
    EVENT_KEY_RELEASE = 4
    EVENT_LEAVE = 5
    EVENT_SCROLL = 6


    # regular expression to test whether a path is absolute
    __IS_ABSOLUTE_PATH_RE = re.compile("[a-zA-Z]+://.+")

    def __init__(self, ident, rep):

        self.__action_stamp = 1
        

        # the display menu
        self.__DISPLAY_MENU = [
            MenuItem("/__cfg", _("_Configure desklet"),
                     callback = self.__handle_configure),
            MenuItem("/__move", _("_Move desklet"),
                     callback = self.__handle_move),
            MenuItem("/__sep1"),
            MenuItem("/__src", _("_View Source"),
                     callback = self.__handle_source),
            MenuItem("/__sep2"),
            MenuItem("/__restart", _("Re_start desklet"),
                     callback = self.__handle_restart),
            MenuItem("/__remove", _("_Remove desklet"),
                     callback = self.__handle_remove)
            ]

        # the layout object
        self.__layout_object = LayoutObject(None)

        # the stack of current mouse cursors
        self.__cursor_stack = []

        # the controls handling the deprecated sensor stuff: id -> control
        self.__sensor_controls = {}

        # mapping of the elements to their parent arrays: child_id -> array_id
        self.__arrays = {}

        # content of this display (valid until display has been initialized)
        self.__content = None

        # scriptlets of this display (valid until display has been initialized)
        self.__scriptlets = []

        # the root TargetGroup
        self.__group = None

        # the scripting environment
        self.__script = Script(ident, rep)

        # the configurator object
        self.__configurator = DisplayConfigger(ident)
        self.__configurator.set_scripting_environment(self.__script)

        # the path of the .display file
        self.__path = os.path.dirname(rep)
        self.__display_file = rep

        # the unique ID of this display
        self.__id = ident

        # the last position of the mouse pointer (used for filtering out
        # unwanted events)
        self.__last_pointer_pos = (-1, -1)

        # mapping between sensors and targets; which target watches
        # which sensor?
        # (sensor, port) -> (target, property)
        self.__mapping = {}

        # temporary data for remembering the position of the last mouse click
        self.__pointer_pos = (0, 0)

        # whether the display reacts on events
        self.__is_sensitive = True


        gtk.HBox.__init__(self)



    #
    # Returns the unique ID of this display.
    #
    def get_id(self): return self.__id

    def get_next_child_index(self): return -1

    def get_index_path(self): return []



    #
    # Initializes the display. Call this after you have finished constructing
    # it.
    #
    def initialize(self):
        assert self.__content

        childtype, settings, children = self.__content

        # load saved positions
        positions = DefaultStateSaver().get_key("positions", {})
        lx, ly = positions.get(self.get_id(), (UNSET_COORD, UNSET_COORD))
        x = settings.get("x", lx)
        y = settings.get("y", ly)
        if (x != UNSET_COORD): settings["x"] = x
        if (y != UNSET_COORD): settings["y"] = y

        # setup the root layout object (representing the root window)
        self.__layout_object.set_geometry(Unit.ZERO, Unit.ZERO,
                              Unit.Unit(gtk.gdk.screen_width(), Unit.UNIT_PX),
                              Unit.Unit(gtk.gdk.screen_height(), Unit.UNIT_PX))

        # build widget hierarchy
        self.new_child(childtype, settings, children)
        self.add(self.__group.get_widget())

        # run initial scripts
        for s in self.__scriptlets:
            self.execute_script(s)

        # load stored configuration
        self.__configurator.load_config()

        # unlock the geometry engine
        gobject.idle_add(self.__group.unlock_geometry)


    #
    # Sets the content of the display.
    #
    def set_content(self, childtype, settings, children):

        self.__content = (childtype, settings, children)



    #
    # Sets metadata.
    #
    def set_metadata(self, author = "", version = "", name= "", preview = ""):

        # a banner without the display's name would look crappy
        if (name):
            if (preview): preview = self.get_full_path(preview)
            self.__configurator.set_banner(preview,
                                  "<big>%s</big> %s\n"
                                  "<small>%s</small>" % (name, version, author))



    #
    # Adds the given scriptlet.
    #
    def add_scriptlet(self, code, filename):

        scriptlet = Scriptlet(code, filename)
        self.__scriptlets.append(scriptlet)



    #
    # Executes the given script.
    #
    def execute_script(self, code):

        self.__script.execute(code)



    #
    # Executes the given callback script.
    #
    def execute_callback_script(self, code, this):

        scriptlet = Scriptlet(code, self.__display_file)
        self.__script.add_element(None, "self", this)
        self.__script.execute(scriptlet)



    #
    # Adds the given target to the scripting environment.
    #
    def add_target_to_script(self, name, target):

        index_path = target.get_index_path()
        length = len(index_path)

        if ("#" in name): name = name[:name.find("#")]
        if (length > 0):
            self.__script.add_element_with_path(_ROOT, name, target, index_path)
        else:
            self.__script.add_element(_ROOT, name, target)



    #
    # Builds the configurator.
    #
    def build_configurator(self, items):

        self.__configurator.build(items)
        self.__configurator.set_path(self.__path)



    #
    # Sends an event to be executed to the display.
    #
    def send_event(self, etype, *args):

        if (not self.__is_sensitive): return
        w, h = self.__group.get_widget().size_request()
        lx, ly = self.__pointer_pos

        self.__action_stamp += 1

        if (etype == self.EVENT_MOTION):
            x, y = args

            utils.request_idle_call(self.__queue_motion, x, y, w, h, False)
            return


        elif (etype == self.EVENT_LEAVE):
            utils.request_idle_call(self.__queue_motion, 0, 0, w, h, True)
            return


        elif (etype == self.EVENT_PRESS):
            button, x, y, counter = args

            self.__pointer_pos = (x, y)
            if (counter == 1):
                action = DisplayTarget.ACTION_PRESS
            else:
                action = DisplayTarget.ACTION_DOUBLECLICK

            event = Struct(button = button, _args = [button])


        elif (etype == self.EVENT_RELEASE):
            button, x, y = args

            if (button == 1):
                if (abs(lx - x) < 10 and abs(ly - y) < 10):
                    action = DisplayTarget.ACTION_CLICK
                else:
                    action = DisplayTarget.ACTION_RELEASE
            elif (button == 2):
                return
            elif (button == 3):
                action = DisplayTarget.ACTION_MENU
            else:
                return
            event = Struct(button = button, _args = [button])


        elif (etype == self.EVENT_SCROLL):
            direction, x, y = args

            if (direction == gtk.gdk.SCROLL_UP):
                direction = 0
            elif (direction == gtk.gdk.SCROLL_DOWN):
                direction = 1
            else:
                direction = -1
            action = DisplayTarget.ACTION_SCROLL
            event = Struct(direction = direction, _args = [direction])


        elif (etype == self.EVENT_KEY_PRESS):
            key, x, y = args
            action = DisplayTarget.ACTION_KEY_PRESS
            event = Struct(key = key)


        elif (etype == self.EVENT_KEY_RELEASE):
            key, x, y = args
            action = DisplayTarget.ACTION_KEY_RELEASE
            event = Struct(key = key)


        else:
            # what kind of event did we get there?
            assert (False)  # never reached

        self.__group.get_layout_object().send_action(Unit.Unit(x, Unit.UNIT_PX),
                                                     Unit.Unit(y, Unit.UNIT_PX),
                                                     self.__action_stamp,
                                                     action, event)
        #self.__group.detect_leave(self.__action_stamp)


        # extend the menu or create one if there's none
        if (action == DisplayTarget.ACTION_MENU):
            utils.request_idle_call(self.open_menu, [])



    #
    # Returns the path of the .display file.
    #
    def get_path(self):

        return self.__path



    #
    # Returns the full path of the given path which may be relative to the
    # .display file.
    #
    def get_full_path(self, path):

        # a path is absolute iff it starts with "/" or with a protocol name
        # such as "http://", otherwise it's relative
        if (path.startswith("/") or self.__IS_ABSOLUTE_PATH_RE.match(path)):
            return path
        else:
            return os.path.join(self.__path, path)



    #
    # Returns the display.
    #
    def _get_display(self): return self



    #
    # Creates and returns a new layout object as a child of this element's
    # layout object.
    #
    def new_layout_object(self):

        return self.__layout_object.new_child()



    def new_child(self, childtype, settings, children):

        import targetregistry
        # we don't catch the KeyError here, but in the DisplayFactory
        try:
            self.__group = targetregistry.create(childtype, self)
        except KeyError, exc:
            log(`exc`)
        self.__group.get_widget().show()
        cid = settings["id"]
        self.add_target_to_script(cid, self.__group)

        for t, s, c in children:
            self.__group.new_child(t, s, c)

        for key, value in settings.items():
            self.__group.set_xml_prop(key, value)



    #
    # Opens the configuration dialog for this display.
    #
    def __open_configurator(self):
        assert (self.__configurator)

        configurators = [s.configurator
                         for s in self.__sensor_controls.values()]

        if (configurators):
            # support old deprecated sensor stuff
            from DisplayConfigurator import DisplayConfigurator
            dconf = DisplayConfigurator(configurators)

        else:
            self.__configurator.show()



    #
    # Saves this display's position.
    #
    def __save_position(self):

        x, y, nil, nil = self.__group.get_user_geometry()
        positions = DefaultStateSaver().get_key("positions", {})
        positions[self.get_id()] = (x.as_px(), y.as_px())
        DefaultStateSaver().set_key("positions", positions)



    #
    # Removes this display.
    #
    def remove_display(self):

        # save the display's position
        self.__save_position()

        self.__configurator.destroy()
        self.update_observer(self.OBS_CLOSED, self.__id)
        self.drop_observers()

        self.__script.stop()
        for s in self.__sensor_controls.values():
            s.stop = True

        self.__group.delete()

        del self.__sensor_controls
        del self.__mapping
        del self.__group



    #
    # Purges this display.
    #
    def purge_display(self):

        # remove the display's position
        positions = DefaultStateSaver().get_key("positions", {})
        try:
            del positions[self.get_id()]
        except KeyError:
            # position wasn't stored
            pass

        # clean up scripting environment
        self.__script.remove()

        # remove the configuration
        self.__configurator.remove_config()



    #
    # Sends the given action with an event object to the display.
    #
    def send_action(self, src, action, event):

        call = src.get_action_call(action)
        if (call):

            # analyze call to see if it's a legacy call
            # FIXME: remove eventually :)
            if (re.match("\w+:.*", call)):
                log("Deprecation: Please use new style call.",
                    is_warning = True)

                try:
                    legacy_args = event._args
                except:
                    legacy_args = []

                legacy_call = actionparser.parse(call)
                path = src.get_index_path()
                self.call_sensor(legacy_call, path, *legacy_args)

            else:
                #src.notify_handle_action(True)
                src._setp("event", event)
                self.execute_callback_script(call, src)



    def __queue_motion(self, px, py, w, h, is_leave):

        # some window managers send a LEAVE event for mouse clicks;
        # work around this
        if (is_leave and (px, py) == self.__last_pointer_pos
            and (0 <= px <= w) and (0 <= py <= h)):
            is_leave = False

        # don't do redundant work
        if (not is_leave and (px, py) == self.__last_pointer_pos): return
        else: self.__last_pointer_pos = (px, py)

        if (not is_leave):
            ux = Unit.Unit(px, Unit.UNIT_PX)
            uy = Unit.Unit(py, Unit.UNIT_PX)
            self.__group.get_layout_object().send_action(ux, uy,
                                                         self.__action_stamp,
                                                   DisplayTarget.ACTION_MOTION,
                                                         Struct())
            #self.__group.handle_action(DisplayTarget.ACTION_MOTION,
            #                           ux, uy, Struct())
            #self.__group.notify_handle_action(True)
            self.__group.detect_leave(self.__action_stamp)
            #self.__action_stamp += 1
        else:
            self.__group.detect_leave(self.__action_stamp)
            #self.__group.notify_handle_action(False)



    #
    # Observer for the root group.
    #
    def child_observer(self, src, cmd):

        if (cmd == src.OBS_GEOMETRY):
            #src.get_layout_object().dump()
            x, y, w, h = self.__group.get_geometry()
            ux, uy = self.__group.get_user_geometry()[:2]

            if (ux.is_unset() or uy.is_unset()):
                utils.request_idle_call(self.update_observer, self.OBS_GEOMETRY,
                                   UNSET_COORD, UNSET_COORD,
                                   w.as_px(), h.as_px())
            else:
                self.__save_position()
                utils.request_idle_call(self.update_observer, self.OBS_GEOMETRY,
                                   x.as_px(), y.as_px(),
                                   w.as_px(), h.as_px())



    #
    # Sets the anchored position of the display.
    #
    def set_position(self, x = UNSET_COORD, y = UNSET_COORD):

        if (x != y != UNSET_COORD):
            x = Unit.Unit(x, Unit.UNIT_PX)
            y = Unit.Unit(y, Unit.UNIT_PX)
            cx, cy, cw, ch = self.__group.get_geometry()

            if ((cx, cy) != (x, y)):
                ax, ay = self.__group.get_anchored_coords(x, y, cw, ch)
                dx, dy= x - ax, y - ay
                new_x = x.as_px() + dx.as_px()
                new_y = y.as_px() + dy.as_px()

                #sx, sy = self.__layout_object.get_scaling()
                self.__group.set_xml_prop("x", str(new_x))# / sx))
                self.__group.set_xml_prop("y", str(new_y))# / sy))



    #
    # Sets the size of the display.
    #
    def set_size(self, width, height):

        self.__group.set_xml_prop("width", str(width))
        self.__group.set_xml_prop("height", str(height))



    #
    # Sets the configuration settings.
    # FIXME: remove eventually :)
    #
    def __set_settings(self, settings, sensor):

        for key, value in settings.get_entries():
            # get all (target, property) tuples that are watching the given
            # sensor key and notify the targets
            entries = self.__mapping.get((sensor, key), [])

            # if an array does not have enough elements, create them
            if (not entries and "[" in key):
                pos = key.find("[")
                port = key[:pos]
                path = key[pos + 1:-1].split("][")
                new_length = int(path.pop()) + 1
                if (port in self.__arrays):
                    array = self.__arrays[port]
                    array.set_prop("length", new_length)
                    entries = self.__mapping.get((sensor, key), [])

            for target, prop in entries:
                target.set_xml_prop(prop, value)



    #
    # Sets the window configuration.
    #
    def set_prop(self, key, value):

        if (key == "window-flags"):
            self.update_observer(self.OBS_FLAGS, value)

        elif (key == "desktop-borders"):
            self.update_observer(self.OBS_BORDERS, value)

        elif (key == "title"):
            self.update_observer(self.OBS_TITLE, value)

        elif (key == "icon"):
            filename = self.get_full_path(value)
            loader = gtk.gdk.PixbufLoader()
            try:
                data = vfs.read_entire_file(filename)
            except:
                return
            try:
                loader.write(data, len(data))
            except:
                log("Warning: Invalid image format.")
                return

            loader.close()
            pixbuf = loader.get_pixbuf()
            self.update_observer(self.OBS_ICON, pixbuf)

        elif (key == "shape"):
            if (value.lstrip().startswith("<")):
                from utils.DOM import DOM
                from utils import svg
                w, h, = self.size_request()
                if (not w or not h): return
                root = DOM(value).get_root()
                root["width"] = `w`
                root["height"] = `h`
                img = gtk.Image()
                svg.render(img, w, h, str(root))
                pixbuf = img.get_pixbuf()

            else:
                filename = self.get_full_path(value)
                loader = gtk.gdk.PixbufLoader()
                try:
                    data = vfs.read_entire_file(filename)
                except:
                    return
                try:
                    loader.write(data, len(data))
                except:
                    log("Warning: Invalid image format.")
                    return

                loader.close()
                pixbuf = loader.get_pixbuf()

            pix, mask = pixbuf.render_pixmap_and_mask(1)
            self.update_observer(self.OBS_SHAPE, mask)



    #
    # Changes the mouse cursor.
    #
    def __set_cursor(self):

        if (self.__cursor_stack):
            cursor = self.__cursor_stack[-1]
        else:
            cursor = None
            
        self.update_observer(self.OBS_CURSOR, cursor)



    #
    # Adds a cursor to the cursor stack.
    #
    def push_cursor(self, cursor):

        self.__cursor_stack.append(cursor.upper())
        utils.request_call(self.__set_cursor)



    #
    # Removes a cursor from the cursor stack.
    #
    def pop_cursor(self, cursor):

        try:
            self.__cursor_stack.remove(cursor.upper())
        except:
            pass
        utils.request_call(self.__set_cursor)
        
        



    #
    # Adds a sensor to this display.
    # FIXME: remove eventually :)
    #
    def add_sensor(self, ident, sensor):

        def set_menu(menu):

            m = []
            cnt = 0
            for entry in menu[:-1]:
                if (not entry):
                    m.append(MenuItem("/sep%d" % cnt))
                else:
                    callback = entry[3]
                    args = [None] + list(entry[4])
                    m.append(MenuItem("/item%d" % cnt, entry[0], None,
                                      callback, args))
                cnt += 1
            #end for

            utils.request_call(self.open_menu, m)
            

        self.__sensor_controls[ident] = sensor
        sensor.bind("output", self.__set_settings, ident)
        sensor.bind("menu", set_menu)



    #
    # Binds a sensor's output to an element.
    # FIXME: remove eventually :)
    #
    def bind_sensor(self, sensorplug, element, prop):

        def h(value, element, prop, port):
            for k, v in value.items():
                if (k == port):
                    element.set_xml_prop(prop, v)

        ident, port = sensorplug.split(":")
        if (not (element, prop) in
              self.__mapping.setdefault((ident, port), [])):
            self.__mapping[(ident, port)].append((element, prop))



    #
    # Calls a function of a Sensor.
    # FIXME: Remove eventually
    #
    def call_sensor(self, cmd, path, *args):
        assert(cmd)

        args = list(args)
        for ident, callname, userargs in cmd:
            sensorctrl = self.__sensor_controls[ident]

            allargs = args + userargs

            # the sensor is an external module, so we make sure it cannot crash
            # the application
            try:
                sensorctrl.action = (callname, path, allargs)

            except StandardError, exc:
                log("The sensor produced an error: %s" % (exc,))



    #
    # Unbinds a sensor's output from an element.
    # FIXME: remove eventually :)
    #
    def unbind_sensor(self, sensorplug, element, prop):

        ident, port = sensorplug.split(":")
        try:
            self.__mapping[(ident, port)].remove((element, prop))
        except KeyError:
            pass



    #
    # Registers an array to be parent array of the given child.
    # FIXME: remove eventually :)
    #
    def register_array_for_port(self, array, sensorplug):

        ident, port = sensorplug.split(":")
        self.__arrays[port] = array



    #
    # Returns the anchored geometry of this display.
    #
    def get_geometry(self):

        try:
            x, y, w, h = self.__group.get_geometry()
            ax, ay = self.__group.get_anchored_coords(x, y, w, h)
            dx, dy = x - ax, y - ay
            return (x + dx, y + dy, w, h)
        except:
            return (Unit.ZERO, Unit.ZERO, Unit.ZERO, Unit.ZERO)



    #
    # Opens a menu.
    #
    def open_menu(self, menu):

        # always work on a copy
        menu = menu[:]
        
        if (menu): menu.append(MenuItem("/__separator"))
        menu += self.__DISPLAY_MENU

        mainmenu = gtk.Menu()
        tree = {}
        for entry in menu:
            if (not entry.label):
                item = gtk.SeparatorMenuItem()

            elif (entry.icon):
                item = gtk.ImageMenuItem(entry.label)
                try:
                    img = Tiling()
                    data = vfs.read_entire_file(self.get_full_path(entry.icon))
                    img.set_from_data(data)
                    img.render(16, 16, 1, 1)
                    img.show()
                    item.set_image(img)
                except:
                    import traceback; traceback.print_exc()
                    pass
                
            else:
                item = gtk.MenuItem(entry.label)

            if (entry.callback):
                item.connect("activate",
                             lambda src, cb, args: cb(*args),
                             entry.callback, entry.args)
                
            item.show()
            if (not entry.active): item.set_sensitive(False)

            menupath = "/".join(entry.path.split("/")[:-1])
            if (menupath):
                parentitem = tree.get(menupath)
                
                if (not parentitem.get_submenu()):
                    m = gtk.Menu()
                    parentitem.set_submenu(m)
                else:
                    m = parentitem.get_submenu()

            else:
                m = mainmenu
            
            m.append(item)
            tree["/".join(entry.path.split("/"))] = item
        #end for

        mainmenu.popup(None, None, None, 0, 0)
            


    def __handle_configure(self, *args):

        self.__open_configurator()



    def __handle_move(self, *args):

        self.update_observer(self.OBS_GEOMETRY, UNSET_COORD, UNSET_COORD, 0, 0)



    def __handle_source(self, *args):

        from config import settings

        if (HAVE_WIN32):
            os.system("start notepad \"%s\"" % (self.__display_file,))
        else:
            os.system("%s \"%s\" & disown" % (settings.editor,
                                              self.__display_file))


    def __handle_restart(self, *args):

        self.update_observer(self.OBS_RESTART, self.__id)



    def __handle_remove(self, *args):

        self.close()



    def close(self):

        def remove_display(*args):
            self.update_observer(self.OBS_CLOSE, self.__id)


        utils.dialog.question(None,
                              _("Do you really want to remove this desklet ?"),
                              _("This desklet will no longer be displayed "
                                "and its configuration will be purged."),
                              (gtk.STOCK_CANCEL, Null),
                              (gtk.STOCK_DELETE, remove_display)
                             )



    #
    # Returns a unique ID string.
    #
    def make_id():
        return "id%d%d" % (int(time.time() * 100), random.randrange(0xffff))

    make_id = staticmethod(make_id)



class __Null:

    """Null object that does __nothing__"""

    def __init__(self, *args, **kwargs): pass
    def __call__(self, *args, **kwargs): return self
    def __repr__(self): return 'Null'
    def __str__(self): return 'Null'
    def __nonzero__(self): return 0

    def __getattr__(self, name): return self
    def __setattr__(self, name, value): return self
    def __delattr__(self, name): return self


# Defines Null as a builtin
# we don't need more than one instance
import __builtin__
__builtin__.Null = __Null()



from GlassWindow import GlassWindow
from config.StateSaver import DefaultStateSaver
from utils.WindowSnapper import WindowSnapper
from layout import Unit
import utils
from main import ICON, UNSET_COORD

import gobject
import gtk


# the available window flags
_WINDOW_FLAG_NONE =      0
_WINDOW_FLAG_ABOVE =     1 << 0
_WINDOW_FLAG_BELOW =     1 << 1
_WINDOW_FLAG_STICKY =    1 << 2
_WINDOW_FLAG_MANAGED =   1 << 3
_WINDOW_FLAG_DECORATED = 1 << 4

# cursor mappings
_CURSORS = {
    "ARROW_RIGHT":       gtk.gdk.ARROW,
    "CROSSHAIR":         gtk.gdk.CROSSHAIR,
    "EDGE_N":            gtk.gdk.TOP_SIDE,
    "EDGE_S":            gtk.gdk.BOTTOM_SIDE,
    "EDGE_W":            gtk.gdk.LEFT_SIDE,
    "EDGE_E":            gtk.gdk.RIGHT_SIDE,
    "EDGE_NW":           gtk.gdk.TOP_LEFT_CORNER,
    "EDGE_NE":           gtk.gdk.TOP_RIGHT_CORNER,
    "EDGE_SW":           gtk.gdk.BOTTOM_LEFT_CORNER,
    "EDGE_SE":           gtk.gdk.BOTTOM_RIGHT_CORNER,
    "FLEUR":             gtk.gdk.FLEUR,
    "KILL":              gtk.gdk.PIRATE,
    "LINK":              gtk.gdk.HAND2,
    "POINTER_1":         gtk.gdk.LEFT_PTR,
    "POINTER_2":         gtk.gdk.RIGHT_PTR,
    "QUESTION":          gtk.gdk.QUESTION_ARROW,
    "TEXT":              gtk.gdk.XTERM,
    "WAIT":              gtk.gdk.WATCH,
    "X":                 gtk.gdk.X_CURSOR
    }


#
# Class for display windows.
#
class Window(GlassWindow):

    # mapping: str -> window flag
    __WINDOW_FLAGS = {"below"  : _WINDOW_FLAG_BELOW,
                      "above"  : _WINDOW_FLAG_ABOVE,
                      "sticky" : _WINDOW_FLAG_STICKY,
                      "managed": _WINDOW_FLAG_MANAGED,
                      "decorated": _WINDOW_FLAG_DECORATED}

    __window_snapper = WindowSnapper()

    def __init__(self, display):

        # window bbox as it was stored in the window snapper
        self.__window_bbox = (0, 0, 0, 0)

        # the last mouse pointer position
        self.__last_pointer = (0, 0)

        # window position for detecting moves
        self.__window_pos = (UNSET_COORD, UNSET_COORD)

        # window size for detecting resizing
        self.__window_size = (0, 0)

        # the window flags
        self.__window_flags = _WINDOW_FLAG_NONE

        # the desktop borders
        self.__desktop_borders = (Unit.Unit(), Unit.Unit())

        # the current mouse cursor
        self.__cursor = None

        # whether we are in managed mode (window manager takes care of the
        # window); managed windows don't snap to desklets and cannot be moved
        # with the middle mouse button
        self.__managed_mode = False

        # temporary data used for dragging windows
        self.__is_dragging = False
        self.__drag_offset = (0, 0)
        self.__dont_snap = False

        self.__shape = None

        GlassWindow.__init__(self, gtk.WINDOW_TOPLEVEL)

        self.set_size_request(-1, -1)
        self.set_default_size(10, 10)

        # set the icon
        self.set_icon(gtk.gdk.pixbuf_new_from_file(ICON))

        self.__display = display


        display.show()
        self.add(display)
        display.add_observer(self.__on_observe_display)

        # set up event handlers
        self.connect("delete-event", self.__on_close)
        self.connect("key-press-event", self.__on_key, 0)
        self.connect("key-release-event", self.__on_key, 1)
        self.connect("button-press-event", self.__on_button, 0)
        self.connect("button-release-event", self.__on_button, True)
        self.connect("motion-notify-event", self.__on_motion, False)
        self.connect("leave-notify-event", self.__on_motion, True)
        self.connect("scroll-event", self.__on_scroll)
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK |
                        gtk.gdk.BUTTON_RELEASE_MASK |
                        gtk.gdk.LEAVE_NOTIFY_MASK |
                        gtk.gdk.POINTER_MOTION_MASK)
                        # POINTER_MOTION_HINT_MASK caused problems...
                        # gtk.gdk.POINTER_MOTION_HINT_MASK)

        self.realize()
        #self.set_window_flags("")
        

    #
    # Observer handler for the display.
    #
    def __on_observe_display(self, src, cmd, *args):

        if (cmd == src.OBS_GEOMETRY):
            x, y, w, h = args
            if (x == UNSET_COORD or y == UNSET_COORD):
                # let the user place the window
                self.__begin_dragging()
            else:
                self.set_position(x, y, True)
                self.set_size(w, h)

        elif (cmd == src.OBS_BORDERS):
            x, y = args[0]
            self.__desktop_borders = (x, y)
            self.__set_struts()

        elif (cmd == src.OBS_FLAGS):
            flags = args[0]
            utils.request_call(self.__set_window_flags, flags)

        elif (cmd == src.OBS_ICON):
            icon = args[0]
            self.set_icon(icon)

        elif (cmd == src.OBS_SHAPE):
            shape = args[0]
            self.__set_shape(shape)

        elif (cmd == src.OBS_TITLE):
            title = args[0]
            self.set_title(title)

        elif (cmd == src.OBS_CURSOR):
            cursor = args[0]
            if (cursor):
                self.__cursor = gtk.gdk.Cursor(_CURSORS.get(cursor))
            else:
                self.__cursor = None

            self.window.set_cursor(self.__cursor)

        elif (cmd == src.OBS_CLOSED):
            wx, wy, w, h = self.__window_bbox
            self.__window_snapper.remove(wx, wy, w, h)
            self.destroy()
                    


    #
    # Digs a hole into the window at the given position. This is necessary to
    # be able to place windows with embedded bonobo controls.
    #
    def __dig_a_hole(self, x, y):

        w, h = self.__window_size
        if (w != 0 and h != 0):
            mask = gtk.gdk.Pixmap(None, w, h, 1)
            gc = mask.new_gc()
            gc.foreground = gtk.gdk.Color(0, 0, 0, 1)
            mask.draw_rectangle(gc, True, 0, 0, w, h)
            gc.foreground = gtk.gdk.Color(0, 0, 0, 0)
            mask.draw_rectangle(gc, True, x - 4 , y - 4, 8, 8)
            self.get_children()[0].window.shape_combine_mask(mask, 0, 0)



    #
    # Begins a window dragging operation.
    #
    def __begin_dragging(self):

        assert(self.window)

        self.__is_dragging = True
        #self.update_observer(self.OBS_LOCK)
        x, y = self.get_pointer()
        w, h = self.__window_size
        w = max(5, w)
        h = max(5, h)
        if (not (0 <= x <= w)): x = w / 2
        if (not (0 <= y <= h)): y = h / 2
        self.__drag_offset = (x, y)
        self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.FLEUR))

        bx, by, bw, bh = self.__window_bbox
        if (not self.__managed_mode):
            self.__window_snapper.remove(bx, by, bw, bh)

        # grab the window focus to ensure that keypresses work
        self.present()

        gobject.timeout_add(50, self.__do_dragging)
        gobject.timeout_add(50, self.__do_move_window)

        return False



    #
    # Ends a window dragging operation.
    #
    def __end_dragging(self):

        assert(self.window)

        self.get_children()[0].window.shape_combine_mask(self.__shape, 0, 0)
        self.__is_dragging = False
        #self.update_observer(self.OBS_UNLOCK)
        self.window.set_cursor(self.__cursor)

        x, y = self.__window_pos
        w, h = self.__window_size
        if (not self.__managed_mode):
            self.__window_snapper.insert(x, y, w, h)

        self.__window_bbox = (x, y, w, h)
        self.__display.set_position(x, y)
        self.__set_struts()

        return False



    #
    # Drags the window.
    #
    def __do_dragging(self):

        if (self.__is_dragging):
            offx, offy = self.__drag_offset
            winx, winy = self.get_position()
            x, y = self.get_pointer()
            rx, ry = winx + x, winy + y

            if ((rx, ry) != self.__last_pointer):
                self.__last_pointer = (rx, ry)
                x += winx; y += winy
            else:
                wx, wy = self.__window_pos
                x = offx + wx; y = offy + wy

            new_x = x - offx
            new_y = y - offy
            if (self.__window_pos != (new_x, new_y)):
                if (not self.__dont_snap and not self.__managed_mode):
                    w, h = self.__window_size
                    new_x, new_y = self.__window_snapper.snap(new_x, new_y,
                                                              w, h)
                self.__window_pos = (new_x, new_y)

            return True

        else:
            return False



    def __do_move_window(self):

        x, y = self.__window_pos
        utils.request_call(self.move, x, y)

        return self.__is_dragging



    def __on_key(self, src, event, is_release):

        key = gtk.gdk.keyval_name(event.keyval)
        if (not is_release and self.__is_dragging):
            x, y = self.__window_pos

            # let the user move the display window with the keyboard
            if (key == "Left"): x -= 4
            elif (key == "Right"): x += 4
            elif (key == "Up"): y -= 4
            elif (key == "Down"): y += 4

            # stop dragging if the Return key is pressed
            elif (key == "Return"):
                self.__end_dragging()

            self.__window_pos = (x, y)

        if (not is_release):
            # don't snap while one of the SHIFT keys is pressed
            self.__dont_snap = (key in ("Shift_L", "Shift_R"))
        else:
            self.__dont_snap = False

        if (not self.__is_dragging):
            px, py = self.get_pointer()
            if (is_release):
                self.__display.send_event(self.__display.EVENT_KEY_RELEASE,
                                          key, px, py)
            else:
                self.__display.send_event(self.__display.EVENT_KEY_PRESS,
                                          key, px, py)
                                          


    def __on_motion(self, src, event, is_leave):

        if (self.__is_dragging): return

        if (is_leave):
            utils.request_call(self.__display.send_event,
                               self.__display.EVENT_LEAVE)
        else:
            utils.request_call(self.__display.send_event,
                               self.__display.EVENT_MOTION, event.x, event.y)



    def __on_button(self, src, event, is_release):

        if (not is_release and event.button == 2 and not self.__managed_mode):
            self.__begin_dragging()

        elif (not is_release and self.__is_dragging):
            self.__end_dragging()

        elif (is_release and self.__is_dragging):
            self.__end_dragging()

        px, py = self.get_pointer()
        if (is_release):
            self.__display.send_event(self.__display.EVENT_RELEASE,
                                      event.button, px, py)
        else:
            if (event.type == gtk.gdk._2BUTTON_PRESS):
                counter = 2
            else:
                counter = 1
                
            self.__display.send_event(self.__display.EVENT_PRESS,
                                      event.button, px, py, counter)


    def __on_scroll(self, src, event):

        px, py = self.get_pointer()
        self.__display.send_event(self.__display.EVENT_SCROLL, event.direction,
                                  px, py)


    def __on_close(self, src, *args):

        self.__display.close()
        return True




    #
    # Sets the position of the window. If dont_set is True, the display will not
    # be notified about the movement.
    #
    def set_position(self, x, y, dont_set = False):

        if (self.__window_pos != (x, y) and not self.__is_dragging):
            utils.request_call(self.move, x, y)
            self.__window_pos = (x, y)
            if (not dont_set): self.__display.set_position(x, y)

            bx, by, bw, bh = self.__window_bbox
            if (not self.__managed_mode):
                self.__window_snapper.remove(bx, by, bw, bh)
            w, h = self.__window_size
            if (not self.__managed_mode):
                self.__window_snapper.insert(x, y, w, h)

            self.__window_bbox = (x, y, w, h)
            self.__set_struts()
            gobject.idle_add(self.show)




    #
    # Sets the size of the window.
    # 
    def set_size(self, width, height):

        if ((width, height) != self.__window_size):
            bx, by, bw, bh = self.__window_bbox

            if (not self.__managed_mode):
                self.__window_snapper.remove(bx, by, bw, bh)

            self.__window_size = (width, height)
            self.resize(width, height)

            if (not self.__managed_mode):
                self.__window_snapper.insert(bx, by, width, height)

            self.__window_bbox = (bx, by, width, height)
            self.__set_struts()
            



    #
    # Sets the window flags.
    #
    def __set_window_flags(self, value):

        is_managed = self.__managed_mode
        flags = _WINDOW_FLAG_NONE
        for p in value:
            p = p.strip()

            try:
                flags |= self.__WINDOW_FLAGS[p]
            except KeyError:
                log("Warning: %s is not a valid window flag." % (p,))

        self._set_flag_below(flags & _WINDOW_FLAG_BELOW)
        self._set_flag_above(flags & _WINDOW_FLAG_ABOVE)
        self._set_flag_sticky(flags & _WINDOW_FLAG_STICKY)
        self._set_flag_managed(flags & _WINDOW_FLAG_MANAGED)
        self._set_flag_decorated(flags & _WINDOW_FLAG_DECORATED)
        self.__window_flags = flags

        # remove from window snapper
        if (not is_managed and flags & _WINDOW_FLAG_MANAGED):
            self.__managed_mode = True
            bx, by, bw, bh = self.__window_bbox
            self.__window_snapper.remove(bx, by, bw, bh)

        # insert into window snapper
        elif (is_managed and not flags & _WINDOW_FLAG_MANAGED):
            self.__managed_mode = False
            bx, by, bw, bh = self.__window_bbox
            self.__window_snapper.insert(bx, by, bw, bh)



    #
    # Sets the window's shape.
    #
    def __set_shape(self, mask):

        self.__shape = mask
        if (not self.__is_dragging):
            self.shape_combine_mask(mask, 0, 0)



    #
    # Sets desktop struts.
    #
    def __set_struts(self):

        border_x, border_y = self.__desktop_borders
        bx, by, bw, bh = self.__window_bbox
        border_x.set_100_percent(bw)
        border_y.set_100_percent(bh)
        b_x = border_x.as_px()
        b_y = border_y.as_px()
        
        x, y, w, h = self.__window_bbox

        if (not border_x.is_unset()):
            deskwidth = gtk.gdk.screen_width()
            leftstrut = x + w
            rightstrut = deskwidth - x
            left = (leftstrut < rightstrut) and leftstrut + b_x or 0
            right = (leftstrut > rightstrut) and rightstrut + b_x or 0
        else:
            left = right = 0

        if (not border_y.is_unset()):
            deskheight = gtk.gdk.screen_height()
            topstrut = y + h
            bottomstrut = deskheight - y
            top = (topstrut < bottomstrut) and topstrut + b_y or 0
            bottom = (topstrut > bottomstrut) and bottomstrut + b_y or 0
        else:
            top = bottom = 0

        self.window.property_change("_NET_WM_STRUT", "CARDINAL", 32,
                             gtk.gdk.PROP_MODE_REPLACE,
                             [left, right, top, bottom])



from config import settings
from utils import wallpaper
from utils.BGWatcher import BGWatcher
from utils.tiling import Tiling

try:
    from utils import ewmh
except ImportError:
    log("Couldn't load EWMH support. Some window flags won't work.")
    ewmh = Null

import gobject
import gtk

_BGWATCHER = BGWatcher()

_MOOSE_DELAY = 100


# this gives us support for real transparency when using a composition manager
_TRANSLUCENT_CMAP = None
if (not HAVE_WIN32):
    # get the ARGB visual for an alpha channel
    visual = gtk.gdk.visual_get_best_with_depth(32)
    if (visual):
        _TRANSLUCENT_CMAP = gtk.gdk.Colormap(visual, True)


#
# Class for transparent windows.
#
class GlassWindow(gtk.Window):

    #
    # Constructor.
    #
    def __init__(self, wintype = gtk.WINDOW_TOPLEVEL):

        # handler for transparency updates
        self.__update_handler = None

        # remember the current window position to detect movements
        self.__position = (0, 0)
        self.__size = (0, 0)


        # push the tranlucent colormap for creating the window and pop it
        # afterwards to not mess up other windows
        if (settings.translucent and _TRANSLUCENT_CMAP):
            gtk.widget_push_colormap(_TRANSLUCENT_CMAP)
        gtk.Window.__init__(self, wintype)
        if (settings.translucent and _TRANSLUCENT_CMAP):
            gtk.widget_pop_colormap()


        self.__layout = gtk.Fixed()
        self.__layout.show()
        gtk.Window.add(self, self.__layout)



        # welcome to the future!
        if (not settings.translucent and not HAVE_WIN32):
            # mommy, my xserver doesn't have translucency... :(
            self.__background = Tiling()
            self.__background.show()
            self.__layout.put(self.__background, 0, 0)
            
            self.connect("configure-event", self.__on_configure)

            _BGWATCHER.add_observer(self.__bg_observer)


        else:
            # real men don't need wimpy pseudo-transparency... 8)
            self.modify_bg(gtk.STATE_NORMAL,
                           gtk.gdk.color_parse("#000000"))

            if (HAVE_WIN32):
                import win32gui
                import winxpgui
                LWA_COLORKEY = 0x01
                LWA_ALPHA = 0x02
                WS_EX_LAYERED = 0x80000
                GWL_EXSTYLE = -20

                self.realize()
                style = win32gui.GetWindowLong(self.window.handle, GWL_EXSTYLE)
                style |= WS_EX_LAYERED
                win32gui.SetWindowLong(self.window.handle, GWL_EXSTYLE, style)
                winxpgui.SetLayeredWindowAttributes(self.window.handle,
                                                    0x0a0a0a, 0xff,
                                                    LWA_COLORKEY)

    #
    # Override add method.
    #
    def add(self, widget):

        self.__layout.put(widget, 0, 0)


    #
    # Override resize method.
    #
    def resize(self, width, height):

        self.__layout.set_size_request(width, height)            
        gobject.idle_add(gtk.Window.resize, self, width, height)



    #
    # Observer method for the background.
    # Connect this method to the BG watcher.
    #
    def __bg_observer(self, src, cmd):

        self.__update_bg()



    #
    # Updates the background for transparency.
    #
    def __update_bg(self):
        
        if (self.__update_handler):
            gobject.source_remove(self.__update_handler)
            self.__update_handler = gobject.timeout_add(_MOOSE_DELAY,
                                                        self.__updater)
        else: self.__updater()



    def __updater(self):

        if (not self.window): return

        x, y = self.window.get_origin()
        width, height = self.window.get_size()
        if (width > 1 and height > 1):
            self.__capture_bg(x, y, width, height)



    #
    # Captures the background to create transparency.
    #
    def __capture_bg(self, x, y, width, height):

        wallpaper.get_wallpaper(self.__background, x, y, width, height)
        self.queue_draw()



    #
    # Reacts on moving the window.
    #
    def __on_configure(self, src, event):

        pos = self.window.get_origin()
        size = self.window.get_size()
        if (pos != self.__position or size != self.__size):
            self.__position = pos
            self.__size = size
            self.__update_bg()



    #
    # Sets the BELOW window flag.
    #
    def _set_flag_below(self, value):

        if (not self.__get_window_manager() in ["Enlightenment"] and
            not self.get_property("visible")):
            gobject.timeout_add(500, self._set_flag_below, value)

        elif (self.window):
            ewmh.set_above(self.window, not value)
            ewmh.set_below(self.window, value)



    #
    # Sets the ABOVE window flag.
    #
    def _set_flag_above(self, value):

        if (not self.__get_window_manager() in ["Enlightenment"] and
            not self.__get_window_manager().startswith("Xfwm4") and
            not self.get_property("visible")):
            gobject.timeout_add(500, self._set_flag_above, value)

        elif (self.window):
            ewmh.set_below(self.window, not value)
            ewmh.set_above(self.window, value)



    #
    # Sets the STICKY window flag.
    #
    def _set_flag_sticky(self, value):

        if (value): self.stick()
        else: self.unstick()



    def _set_type_hint_dock(self, window, value):

        ewmh.set_type_dock(window, value)

        #value and window.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_DOCK) or \
        #          window.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_NORMAL)



    #
    # Sets the MANAGED window flag.
    #
    def _set_flag_managed(self, value):

        if (value):
            self.set_property("skip-taskbar-hint", 0)
            self.set_property("skip-pager-hint", 0)
            self._set_type_hint_dock(self.window, False)
        else:
            self.set_property("skip-taskbar-hint", 1)
            self.set_property("skip-pager-hint", 1)
            wm = self.__get_window_manager()
            if (wm == "Metacity"): self._set_type_hint_dock(self.window, True)



    #
    # Sets the DECORATED window flag.
    #
    def _set_flag_decorated(self, value):

        if (value): self.set_decorated(True)
        else: self.set_decorated(False)



    #
    # Returns the name of the running EWMH compliant window manager or "".
    #
    def __get_window_manager(self):

        if (HAVE_WIN32): return "Windows"

        name = ""
        win = _WM_NAME_WIN
        if (win != None):
            try:
                name = win.property_get("_NET_WM_NAME")[2]
            except TypeError, exc:
                log("Your window manager doesn't support _NET_WM_NAME!\n"
                    "Switch to a EWMH compliant WM.\n"
                    "The following error occured:\n%s" % (exc,))
                return name

        return name



if (not HAVE_WIN32):
    # get the window where the EMWH compliant window manager tells its name
    root = gtk.gdk.get_default_root_window()

    try:
        ident = root.property_get("_NET_SUPPORTING_WM_CHECK", "WINDOW")[2]
        _WM_NAME_WIN = gtk.gdk.window_foreign_new(long(ident[0]))
    except TypeError, exc:
        _WM_NAME_WIN = ""
        log("Your window manager doesn't support "
            "_NET_SUPPORTING_WM_CHECK! Switch to a compliant WM!"
            "The following error occured:\n%s" % (exc,))




"""
This module holds the datatypes known to gDesklets.
Datatypes are structs containing the type name and some functions for working
with types.

This module further contains some helper functions for dealing with datatypes.

A datatype consists of:
  - name
  - type checker function to check the type of the given value
  - representation function to get a storable representation of the given value
  - build function to build a value from the given representation

"""



TYPE_UNKNOWN = (
    "unknown",                              # name
    lambda v: True,                         # type checker
    lambda v: repr(v),                      # get representation
    lambda r: r                             # build from representation
    )

TYPE_ANY = (
    "any",
    lambda v: True,
    lambda v: repr(v),
    lambda r: r
    )

TYPE_STRING = (
    "string",
    lambda v: type(v) in (type(""), type(u"")),
    lambda v: v,
    lambda r: r
)

TYPE_INT = (
    "integer",
    lambda v: type(v) in (type(1), type(1L)),
    lambda v: str(v),
    lambda r: int(r)
    )

TYPE_FLOAT = (
    "float",
    lambda v: type(v) in (type(0.5), type(1), type(1L)),
    lambda v: str(v),
    lambda r: float(r)
    )

TYPE_LIST = (
    "list",
    lambda v: type(v) in (type([]), type(())),
    lambda v: repr(v),
    lambda r: list(v)
    )

TYPE_BOOL = (
    "bool",
    lambda v: type(v) in (type(True), type(1), type(1L)),
    lambda v: v,
    lambda r: r
    )

from layout.Unit import Unit as _Unit
TYPE_UNIT = (
    "unit",
    lambda v: isinstance(v, _Unit),
    lambda v: repr(v),
    lambda r: _Unit(string = r)
    )

TYPE_UNIT_LIST = (
    "unit list",
    lambda v: isinstance(v[0], _Unit),
    lambda v: repr(v),
    lambda r: r
    )

TYPE_SECRET_STRING = (
    "secret string",
    lambda v: type(v) == type("")
    )

TYPE_OBJECT = (
    "object",
    lambda v: True  # anything matches
    )



def dtype_get_type(name):
    """Returns the datatype with the given name or TYPE_UNKNOWN if the type does
       not exist."""

    for dtype in ( TYPE_ANY, TYPE_STRING, TYPE_INT,
                   TYPE_FLOAT, TYPE_LIST, TYPE_BOOL,
                   TYPE_UNIT, TYPE_UNIT_LIST ):
        if (dtype_name(dtype) == name):
            return dtype
    #end for

    return TYPE_UNKNOWN


def dtype_name(dtype): return dtype[0]
def dtype_check(dtype, v): return dtype[1](v)
def dtype_repr(dtype, v): return dtype[2](v)
def dtype_build(dtype, r): return dtype[3](r)


def dtype_guess(value):
    """Guesses and returns the datatype of the given value."""

    for dtype in ( TYPE_UNIT, TYPE_LIST, TYPE_BOOL,
                   TYPE_STRING, TYPE_INT, TYPE_FLOAT,
                   TYPE_UNKNOWN ):
        if (dtype_check(dtype, value)): return dtype

    #end for

    # never reached since TYPE_UNKNOWN always matches
    assert False



from shell.Plugin import Plugin
from shell.ItemBrowser import ItemBrowser
from plugin.Interface import Interface
from ControlInspector import ControlInspector
from main import HOME

import gtk
import gobject
import os
import textwrap
import string



class UI_ControlBrowser(Plugin):

    def init(self):

        self.__collection = self._get_plugin("Core_ControlCollection")
        self.__browser = ItemBrowser(self.__search_handler,
                                     self.__item_renderer,
                                     gtk.gdk.Pixbuf, gobject.TYPE_STRING,
                                     gobject.TYPE_STRING)
        self.__browser.add_observer(self.__on_activate)

        for name, items in \
           [(_("by interface"), self.__get_interfaces(self.__collection)),
          (_("alphabetically"), self.__get_alphabetically(self.__collection))]:
            self.__browser.add_category(name, name, items)


        shell = self._get_plugin("UI_Shell")

        w1, w2 = self.__browser.get_widgets()
        shell.set_sidebar(w1)
        shell.set_body(w2)



    def show(self):

        w1, w2 = self.__browser.get_widgets()
        w1.show_all()
        w2.show_all()


    def hide(self):

        w1, w2 = self.__browser.get_widgets()
        w1.hide()
        w2.hide()


    def set_location(self, expr):

        self.__browser.set_query(expr)



    def __on_activate(self, src, cmd, item):

        if (item):
            ci = ControlInspector(item)
            ci.show()
    

    def __search_handler(self, expr):

        status = self._get_plugin("Shell_StatusBar")
        try:
            matches = self.__collection.query(expr)
            status.set_status(_("Found %d controls.") % (len(matches)))
            matches.sort(lambda a,b: cmp(a.__name__, b.__name__))
        except:
            status.set_status(_("Invalid search expression."))
            matches = None
            
        return matches



    def __item_renderer(self, item):

        icon = os.path.join(self._get_path(), "control.png")
        pixbuf = gtk.gdk.pixbuf_new_from_file(icon)
        name = item.__name__
        interfaces = Interface.get_interfaces(item)
        ifaces = []
        for i in interfaces:
            ifaces.append(Interface.get_id(i))

        return (pixbuf, name, "\n".join(ifaces))



    def __get_interfaces(self, collection):

        icon = os.path.join(self._get_path(), "interface.png")
        values = []
        matches = collection.query("(MATCH 'interface' '*')")
        for m in matches:
            values += [ (icon, Interface.get_id(iface),
                         "(MATCH 'interface' '%s')" % (Interface.get_id(iface)))
                        for iface in Interface.get_interfaces(m) ]

        return values



    def __get_alphabetically(self, collection):

        icon = os.path.join(self._get_path(), "folder.png")
        
        values = []
        
        for c in string.ascii_uppercase:
            query = "(MATCH 'name' '%c*')" % (c,)
            if collection.query( query ):
                values.append((icon, c, query))

        return values

        

def get_class(): return UI_ControlBrowser



from ConfigDialog import ConfigDialog
from StateSaver import StateSaver
from utils.datatypes import *


class DisplayConfigger(ConfigDialog):
    """
      Configuration Dialog for the displays. This class handles loading / saving
      the configuration.
    """

    def __init__(self, ident):

        # init the StateSaver
        self.__backend = StateSaver(ident + "CONFIG")

        # may we save already, or is it too early?
        self.__may_save = False

        # ID for storing the configuration
        self.__ident = ident

        # the scripting environment with which the config interacts
        self.__scripting_environment = None

        ConfigDialog.__init__(self)
        self.set_property("title", _("Configuration"))



    #
    # Sets the scripting environment to be used.
    #
    def set_scripting_environment(self, script):

        self.__scripting_environment = script
        self._set_setter(self.__setter_wrapper)
        self._set_getter(self.__scripting_environment.get_value)
        self._set_caller(self.__scripting_environment.call_function)



    #
    # Wraps the config setter handler to include saving config.
    #
    def __setter_wrapper(self, key, value, datatype):
        assert key

        self.__scripting_environment.set_value(key, value)

        # save config, but not too early
        if (self.__may_save):
            rep = dtype_repr(datatype, value)
            self.__backend.set_key(key, (rep, dtype_name(datatype)))



    #
    # Build preferences and put elements into the scripting environment
    #
    def build(self, items):

        ConfigDialog.build(self, items)

        for c in self.get_config_items():
            ident = c.get_prop("id")
            if (ident):
                self.__scripting_environment.add_element("Prefs", ident, c)



    #
    # Loads the initial configuration.
    #
    def load_config(self):

        # read stored configuration
        for key in self.__backend.list():
            if (key.endswith("_TYPE")): continue

            rep, dtype = self.__backend.get_key(key)
            datatype = dtype_get_type(dtype)
            value = dtype_build(datatype, rep)

            try:
                self.__setter_wrapper(key, value, None)
            except:
                log("Couldn't pass arguments (%s, %s) to setter."
                    % (key, value))

        # have the children update themselves
        for c in self.get_config_items(): c.update()



    #
    # Removes the configuration.
    #
    def remove_config(self):

        self.__backend.remove()



    def show(self):

        self.__may_save = True
        ConfigDialog.show(self)



import gtop
from Device import Device

import time
import threading

class Net(Device):

    """Provides storage for interface statistics in order to poll
    multiple network devices at the same time"""

    def __init__(self, dev):

        Device.__init__(self)

        self.__dev = dev
        buf = gtop.netload(dev)
        self.__in, self.__out = buf.bytes_in, buf.bytes_out
        self.__time = time.time()
        self.__lock = threading.Lock()


    def poll(self):

        """Returns .
        self.poll() is threadsafe"""

        self.__lock.acquire()
        try:
            buf = gtop.netload(self.__dev)
            bytes_in, bytes_out = buf.bytes_in, buf.bytes_out

            now      = time.time()
            interval = now - self.__time

            in_diff  = bytes_in  - self.__in
            out_diff = bytes_out - self.__out

            speed_in  = int(in_diff / interval)
            speed_out = int(out_diff / interval)

            self.__time = now
            self.__in, self.__out = bytes_in, bytes_out

            return (speed_in, speed_out)

        finally:
            self.__lock.release()



from shell.Plugin import Plugin
from main import CONTROLPATHS
from plugin.PluginRegistry import PluginRegistry


#
# Plugin for the collection of installed controls.
#
class Core_ControlCollection(Plugin):

    def init(self):
        self.__registry = PluginRegistry(*CONTROLPATHS)


    def query(self, query):

        matches = self.__registry.get_plugins(query)
        return matches


def get_class(): return Core_ControlCollection



from plugin.Interface import Interface
from plugin import Permission


class ISensor(Interface):

    sensor = Permission.WRITE
    action = Permission.WRITE
    config_id = Permission.WRITE
    stop = Permission.WRITE
    output = Permission.READ
    menu = Permission.READ
    configurator = Permission.READ



from shell.Plugin import Plugin

import gtk


#
# View for browsing the installed desklets.
#
class View_Displays(Plugin):

    def activate(self):

        #self._get_plugin("UI_Bookmarks").show()
        self._get_plugin("UI_DisplayBrowser").show()


    def deactivate(self):

        #self._get_plugin("UI_Bookmarks").hide()
        self._get_plugin("UI_DisplayBrowser").hide()


    def get_name(self):

        return "Displays"



    def get_icon(self):

        img = gtk.Image()
        img.set_from_stock(gtk.STOCK_NEW, gtk.ICON_SIZE_SMALL_TOOLBAR)
        img.show()
        
        return img


    def get_priority(self): return 0
    

def get_class(): return View_Displays
